#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Final, List, Dict
from typing_extensions import Self

from ....codable import Codable, KeyDescriptor


class BBox(Codable):

    """
        Bounding Box as a python class with utility methods

        Properties
        ----------
        minX : int
            top left x coordinate
        minY : int
            top left y coordinate
        width : int
            width of the bounding box
        height : int
            height of the bounding box
    """

    def __init__(self, minX: int = 0, minY: int = 0, width: int = 0, height: int = 0) -> None:
        self.minX: Final = minX
        self.minY: Final = minY

        self.width: Final = width
        self.height: Final = height

    @property
    def maxX(self) -> int:
        """
            Returns
            -------
            int -> bottom right x coordinate
        """

        return self.minX + self.width

    @property
    def maxY(self) -> int:
        """
            Returns
            -------
            int -> bottom right y coordinate
        """

        return self.minY + self.height

    @property
    def polygon(self) -> List[int]:
        """
            Returns
            -------
            List[int] -> Bounding box represented as a polygon (x, y) values
        """

        return [
            self.minX, self.minY,  # top left
            self.maxX, self.minY,  # top right
            self.maxX, self.maxY,  # bottom right
            self.minX, self.maxY,  # bottom left
            self.minX, self.minY   # top left
        ]

    @property
    def area(self) -> int:
        return self.width * self.height

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["minX"] = KeyDescriptor("top_left_x")
        descriptors["minY"] = KeyDescriptor("top_left_y")

        return descriptors

    @classmethod
    def create(cls, minX: int, minY: int, maxX: int, maxY: int) -> Self:
        """
            Utility constructor which has maxX and maxY as parameters instead
            of width and height

            Parameters
            ----------
            minX : int
                top left x coordinate
            minY : int
                top left y coordinate
            maxX : int
                bottom right x coordinate
            maxY : int
                bottom right y coordinate

            Returns
            -------
            Self -> bounding box
        """

        return cls(minX, minY, maxX - minX, maxY - minY)

    @classmethod
    def fromPoly(cls, polygon: List[int]) -> Self:
        """
            Creates bounding box from a polygon, by finding
            the minimum x and y coordinates and calculating
            width and height of the polygon

            Parameters
            ----------
            polygon : List[int]
                list of x, y points - length must be even

            Returns
            -------
            Self -> bounding box

            Example
            -------
            >>> from coretex import Bbox
            \b
            >>> polygon = [0, 0, 0, 3, 4, 3, 4, 0]
            >>> bbox = Bbox.fromPoly(polygon)
            >>> print(f"minX: {bbox.minX}, minY: {bbox.minY}, width: {bbox.width}, height: {bbox.height}")
            "minX: 0, minY: 0, width: 4, height: 3"
        """

        x: List[int] = []
        y: List[int] = []

        for index, value in enumerate(polygon):
            if index % 2 == 0:
                x.append(value)
            else:
                y.append(value)

        return cls.create(min(x), min(y), max(x), max(y))

    def iou(self, other: 'BBox') -> float:
        """
            Calculate Intersection over Union (IoU) between two bounding boxes

            Parameters:
            other : BBox
                bounding box for which the IoU will be calculated

            Returns
            float -> IoU score
        """

        x1 = max(self.minX, other.minX)
        y1 = max(self.minY, other.minY)
        x2 = min(self.maxX, other.maxX)
        y2 = min(self.maxY, other.maxY)

        intersectionArea = max(0, x2 - x1) * max(0, y2 - y1)

        unionArea = self.area + other.area - intersectionArea
        return intersectionArea / unionArea if unionArea > 0 else 0.0



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from pathlib import Path
from zipfile import ZipFile

from ..local_sample import LocalSample
from ....utils import file as file_utils


def getSequenceFile(directoryPath: Path, extensions: List[str]) -> Path:
    for path in directoryPath.iterdir():
        if path.suffix in extensions:
            return path

    raise FileNotFoundError(f">> [Coretex] {directoryPath} has no files with extensions \"{extensions}\"")


def getForwardSequenceFile(directoryPath: Path, extensions: List[str]) -> Path:
    for path in directoryPath.iterdir():
        if "_R1_" in path.name and path.suffix in extensions:
            return path

    raise FileNotFoundError(f">> [Coretex] {directoryPath} has no files with \"_R1_\" in name and extensions \"{extensions}\"")


def getReverseSequenceFile(directoryPath: Path, extensions: List[str]) -> Path:
    for path in directoryPath.iterdir():
        if "_R2_" in path.name and path.suffix in extensions:
            return path

    raise FileNotFoundError(f">> [Coretex] {directoryPath} has no files with \"_R2_\" in name and extensions \"{extensions}\"")


class LocalSequenceSample(LocalSample):

    """
        Represents the local custom Sample class
        which is used for working with Other Task locally
    """

    @classmethod
    def supportedExtensions(cls) -> List[str]:
        return [".fasta", ".fastq", ".fa", ".fq"]

    @property
    def sequencePath(self) -> Path:
        """
            Returns the path of the .fasta or .fastq sequence file
            contained inside the sample. If the sample contains gzip compressed
            sequences, you will have to call Sample.unzip method first otherwise
            calling Sample.sequencePath will raise an exception

            Raises
            ------
            FileNotFoundError -> if no .fasta, .fastq, .fq, or .fq files are found inside the sample
        """
        return getSequenceFile(self.path, self.supportedExtensions())

    @property
    def forwardPath(self) -> Path:
        """
            Returns the path of the .fasta or .fastq forward sequence file
            contained inside the sample. "_R1_" must be present in the filename
            otherwise it will not be recongnized. If the sample contains gzip compressed
            sequences, you will have to call Sample.unzip method first otherwise
            calling Sample.sequencePath will raise an exception

            Raises
            ------
            FileNotFoundError -> if no .fasta, .fastq, .fq, or .fq files are found inside the sample
        """
        return getForwardSequenceFile(self.path, self.supportedExtensions())

    @property
    def reversePath(self) -> Path:
        """
            Returns the path of the .fasta or .fastq sequence file
            contained inside the sample. "_R2_" must be present in the filename
            otherwise it will not be recongnized. If the sample contains gzip compressed
            sequences, you will have to call Sample.unzip method first otherwise
            calling Sample.sequencePath will raise an exception

            Raises
            ------
            FileNotFoundError -> if no .fasta, .fastq, .fq, or .fq files are found inside the sample
        """
        return getReverseSequenceFile(self.path, self.supportedExtensions())

    def unzip(self, ignoreCache: bool = False) -> None:
        super().unzip(ignoreCache)
        for extension in self.supportedExtensions():
            for compressedSequencePath in self.path.glob(f"*{extension}.gz"):
                decompressedSequencePath = compressedSequencePath.parent / compressedSequencePath.stem

                file_utils.gzipDecompress(compressedSequencePath, decompressedSequencePath)

    def isPairedEnd(self) -> bool:
        """
            This function returns True if the sample holds paired-end reads and
            False if it holds single end. Files for paired-end reads must contain
            "_R1_" and "_R2_" in their names, otherwise an exception will be raised.
            If the sample contains gzip compressed sequences, you will have to call
            Sample.unzip method first otherwise calling Sample.isPairedEnd will
            raise an exception

            Raises
            ------
            FileNotFoundError -> if no files meeting the requirements for either single-end
                or paired-end sequencing reads
        """

        with ZipFile(self.zipPath, "r") as archive:
            sampleContent = archive.namelist()

        for extension in self.supportedExtensions():
            fileNames: List[str] = []

            for fileName in sampleContent:
                if fileName.endswith(extension) or fileName.endswith(f"{extension}.gz"):
                    fileNames.append(fileName)

            if len(fileNames) == 0:
                continue

            if len(fileNames) == 1:
                return False

            if len(fileNames) == 2:
                forwardPresent = any(["_R1_" in fileName for fileName in fileNames])
                reversePresent = any(["_R2_" in fileName for fileName in fileNames])

                if forwardPresent and reversePresent:
                    return True

        raise FileNotFoundError(f">> [Coretex] Invalid sequence sample \"{self.name}\". Could not determine sequencing type")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Literal
from logging import Formatter, LogRecord

import json

from .severity import LogSeverity
from .utils import colorMessage


FormatStyle = Literal["%", "{", "$"]


class CTXFormatter(Formatter):

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: FormatStyle = "%",
        validate: bool = True,
        color: bool = True,
        jsonOutput: bool = True
    ) -> None:

        super().__init__(fmt, datefmt, style, validate)

        self.color = color
        self.jsonOutput = jsonOutput

    def format(self, record: LogRecord) -> str:
        # INFO -> Info, ERROR -> Error, etc...
        record.levelname = record.levelname.lower().capitalize()

        severity = LogSeverity.fromLevel(record.levelno)
        formatted = super().format(record)

        if self.color:
            formatted = colorMessage(severity, formatted)

        if self.jsonOutput:
            return json.dumps({
                "severity": severity.value,
                "message": formatted + "\n"
            })

        return formatted



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Final

from .base_converter import ConverterProcessorType, BaseConverter
from .converters import *


class ConverterProcessorFactory:
    """
        Factory class used to create different types
        of BaseConverter objects based on the convertProcessorType
        property of the class

        Properties
        ----------
        convertProcessorType : ConverterProcessorType
            type of converter for convertin dataset
    """

    def __init__(self, convertProcessorType: ConverterProcessorType):
        self.type: Final = convertProcessorType

    def create(self, datasetName: str, projectId: int, datasetPath: str) -> BaseConverter:
        """
            Creates BaseConverter based on the convertProcessorType
            property of the class

            Parameters
            ----------
            datasetName : str
                name of dataset
            projectId : int
                id of Coretex Project
            datasetPath : str
                path to dataset

            Returns
            -------
            Appropriate BaseConverter object based on
            convertProcessorType
        """

        if self.type == ConverterProcessorType.coco:
            return COCOConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.yolo:
            return YoloConverter(datasetName, projectId, datasetPath)

        if self.type ==  ConverterProcessorType.createML:
            return CreateMLConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.voc:
            return VOCConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.labelMe:
            return LabelMeConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.pascalSeg:
            return PascalSegConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.humanSegmentation:
            return HumanSegmentationConverter(datasetName, projectId, datasetPath)

        if self.type == ConverterProcessorType.cityScape:
            return CityScapeConverter(datasetName, projectId, datasetPath)

        raise RuntimeError()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Tuple, List, Generic
from abc import abstractmethod

from .base_parameter import BaseParameter, T


class BaseListParameter(BaseParameter[T], Generic[T]):

    @property
    def types(self) -> List[type]:
        return [list]

    @property
    @abstractmethod
    def listTypes(self) -> List[type]:
        pass

    def validate(self) -> Tuple[bool, Optional[str]]:
        isValid, error = super().validate()
        if not isValid:
            return isValid, error

        if not self.required and self.value is None:
            return True, None

        if self.required and len(self.value) == 0:  # type: ignore[arg-type]
            return False, f"Required parameter \"{self.name}\" must contain a non-empty array"

        # self.value is of type Optional[Any], but base class validate method already
        # checks if it is a list, if that fails this is not reachable
        for element in self.value:  # type: ignore[union-attr]

            # bool is a subclass of int, do not allow validation to pass if
            # we are looking for integer, but bool is received
            if isinstance(element, bool) and int in self.listTypes and not bool in self.listTypes:
                return False, None

            if not any(isinstance(element, type_) for type_ in self.listTypes):
                return False, None

        return True, None



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from decimal import Decimal, ROUND_HALF_UP


def mathematicalRound(value: float, decimalPlaces: int) -> float:
    """
        Performes mathematical (ROUND_HALF_UP) rounding
        Ex. >= 1.5 will be rounded to 2, < 1.5 will be rounded to 1

        Parameters
        ----------
        value : float
            value to be rounded
        decimalPlaces : int
            amount of decimal places to which the value will be rounded

        Returns
        -------
        float -> the rounded value
    """

    decimal = Decimal(str(value))
    places = Decimal(10) ** -decimalPlaces

    return float(decimal.quantize(places, rounding = ROUND_HALF_UP))


def formatBytes(value: int, precision: int = 2) -> str:
    """
        Formats a size in bytes into a human-readable format
        Ex. value 1444 will be formatted as 1.44 KB

        Parameters
        ----------
        value : int
            size in bytes
        precision : int
            number of decimal places (default is 2)

        Returns
        -------
        str -> formatted string with size and appropriate unit
    """

    suffixes = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    index = 0
    unitBase = 1000
    adjustedValue = float(value)

    while adjustedValue >= unitBase and index < len(suffixes) - 1:
        index += 1
        adjustedValue /= unitBase

    return f"{adjustedValue:.{precision}f} {suffixes[index]}"



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict, Union, Optional
from typing_extensions import Self, override
from datetime import datetime
from zipfile import ZipFile
from pathlib import Path

import json

from ... import folder_manager
from ...networking import networkManager, NetworkObject, ChunkUploadSession, MAX_CHUNK_SIZE, NetworkRequestError
from ...codable import KeyDescriptor


class Model(NetworkObject):

    """
        Represents a machine learning model object on Coretex.ai
        Contains properties that describe the model

        Properties
        ----------
        name : str
            model name
        createdById : str
            id of model
        createdOn : datetime
            date of model creation
        datasetId : int
            dataset id that is used for training the model
        projectId : int
            project id that is used for training the model
        taskId : int
            task id that is used for training the model
        isTrained : bool
            True if model is trained, False otherwise
        isDeleted : bool
            True if model is deleted, False otherwise
        accuracy : float
            model accuracy
        taskRunId : int
            TaskRun id of trained model
        meta : Dict[str, Any]
            model meta data
    """

    name: str
    createdById: str
    createdOn: datetime
    datasetId: int
    projectId: int
    taskId: int
    isTrained: bool
    isDeleted: bool
    accuracy: float
    taskRunId: int
    meta: Dict[str, Any]

    @property
    def path(self) -> Path:
        return folder_manager.modelsFolder / str(self.id)

    @property
    def zipPath(self) -> Path:
        return self.path.with_suffix(".zip")

    @classmethod
    def modelDescriptorFileName(cls) -> str:
        """
            Returns
            -------
            str -> name of model descriptor file
        """

        return "model_descriptor.json"

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["taskRunId"] = KeyDescriptor("model_queue_id")

        return descriptors

    @classmethod
    def createModel(
        cls,
        name: str,
        taskRunId: int,
        accuracy: float,
        meta: Optional[Dict[str, Any]] = None
    ) -> Self:

        """
            Creates Model as a result of TaskRun

            Parameters
            ----------
            name : str
                model name
            taskRunId : int
                TaskRun id of model
            accuracy : float
                model accuracy
            meta : Optional[Dict[str, Any]]
                model metadata

            Returns
            -------
            Self -> Model object

            Raises
            -------
            NetworkRequestError -> If model creation failed

            Example
            -------
            >>> from coretex import Model, currentTaskRun
            >>> model = Model.createModel("model-name", currentTaskRun().id, 0.87)
        """

        if meta is None:
            meta = {}

        return cls.create(
            name = f"{taskRunId}-{name}",
            model_queue_id = taskRunId,
            accuracy = accuracy,
            meta = meta
        )

    @classmethod
    def createProjectModel(
        cls,
        name: str,
        projectId: int,
        accuracy: float,
        meta: Optional[Dict[str, Any]] = None
    ) -> Self:

        """
            Creates Model object inside of the provided Project with specified properties

            Parameters
            ----------
            name : str
                Model name
            projectId : int
                Project to which the Model will be added
            accuracy : float
                Model accuracy
            meta : Dict[str, Any]
                Model metadata

            Returns
            -------
            Self -> Model object

            Example
            -------
            >>> from coretex import Model
            >>> model = Model.createProjectModel("model-name", 123, 0.87)
        """

        if meta is None:
            meta = {}

        return cls.create(
            name = name,
            project_id = projectId,
            accuracy = accuracy,
            meta = meta
        )

    @classmethod
    def saveModelDescriptor(cls, path: Union[Path, str], contents: Dict[str, Any]) -> None:
        """
            Saves a model descriptor - a JSON file that provides a description of a
            machine learning model. It includes information such as the model's
            architecture, input and output shapes, labels, description and etc.

            Parameters
            ----------
            path : Union[Path, str]
                path to where the model descriptor will be saved
            contents : Dict[str, Any]
                key-value pairs which will be stored as json

            Example
            -------
            >>> from coretex import currentTaskRun, Model
            >>> model = Model.createModel("model-name", currentTaskRun().id, accuracy)
            >>> model.saveModelDescriptor(modelPath, {
                    "project_task": currentTaskRun().projectType,
                    "labels": labels,
                    "modelName": model.name,
                    "description": currentTaskRun().description,

                    "input_description":
                        Input shape is [x, y]

                        x is actually number of samples in dataset\n
                        y represents number of unique taxons for selected level in dataset,

                    "input_shape": [x, y],

                    "output_description":
                        Output shape - [x, z]

                        x is actually number of samples in dataset\n
                        z represents that output 2d array (table) is going to have only 1 column (1 prediction for each sample in dataset),

                    "output_shape": [x, z]
                })
        """

        if isinstance(path, str):
            path = Path(path)

        modelDescriptorPath = path / cls.modelDescriptorFileName()

        with modelDescriptorPath.open("w", encoding = "utf-8") as file:
            json.dump(contents, file, ensure_ascii = False, indent = 4)

    @override
    def entityUrl(self) -> str:
        return f'model-item?id={self.id}'

    def download(self, path: Optional[Path] = None, ignoreCache: bool = False) -> None:
        """
            Downloads and extracts the model zip file from Coretex.ai
        """

        if path is None:
            path = self.path

        if self.isDeleted or not self.isTrained:
            return

        if path.exists() and not ignoreCache:
            return

        modelZip = path.with_suffix(".zip")
        response = networkManager.download(f"{self._endpoint()}/download", modelZip, {
            "id": self.id
        })

        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to download Model")

        with ZipFile(modelZip) as zipFile:
            zipFile.extractall(path)

    def upload(self, path: Union[Path, str]) -> None:
        """
            Uploads the provided model folder as zip file to Coretex.ai

            Parameters
            ----------
            path : Union[Path, str]
                Path to the model directory

            Raises
            -------
            ValueError -> if provided path is not a directory
            NetworkRequestError -> if Model upload failed

            Example
            -------
            >>> from coretex import Model, currentTaskRun
            >>> model = Model.createModel("model-name", currentTaskRun().id, 0.87)
            >>> model.upload("path/to/model-dir")
        """

        if isinstance(path, str):
            path = Path(path)

        if not path.is_dir():
            raise ValueError("\"path\" must be a directory")

        zipPath = path.with_suffix(".zip")
        with ZipFile(zipPath, "w") as zipFile:
            for value in path.rglob("*"):
                if not value.is_file():
                    continue

                zipFile.write(value, value.relative_to(path))

        uploadSession = ChunkUploadSession(MAX_CHUNK_SIZE, zipPath)
        uploadId = uploadSession.run()

        parameters = {
            "id": self.id,
            "file_id": uploadId
        }

        response = networkManager.formData("model/upload", parameters)
        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to upload model")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, Dict


def badRequest(error: str) -> Dict[str, Any]:
    """
        Creates a json object for bad request

        Parameters
        ----------
        error : str
            Error message

        Returns
        -------
        dict -> Json object for bad request
    """

    return {
        "code": 400,
        "body": {
            "error": error
        }
    }


def success(data: Optional[Any] = None) -> Dict[str, Any]:
    """
        Creates a json object for successful request

        Parameters
        ----------
        data : Optional[Any]
            Response data

        Returns
        -------
        dict -> Json object for successful request
    """

    if data is None:
        data = {}

    return {
        "code": 200,
        "body": data
    }



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional, Type, Dict, Tuple
from typing_extensions import Self
from datetime import datetime
from enum import Enum
from uuid import UUID

import inflection

from .descriptor import KeyDescriptor
from ..utils.date import DATE_FORMAT, decodeDate


class Codable:

    """
        Class whose subclasses can be serialized/deserialized into/from a JSON format object
    """

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        """
            Defines a custom mapping for python objects to json field
            Key of the dictionary represents a name of the field in python
            Value of the dictionary represents how should the object be serialized/deserialized

            Returns
            -------
            Dict[str, KeyDescriptor] -> Dictionary of objects describing translation
        """

        return {}

    @classmethod
    def __keyDescriptorByJsonName(cls, jsonName: str) -> Tuple[Optional[str], Optional[KeyDescriptor]]:
        for key, value in cls._keyDescriptors().items():
            if value.jsonName == jsonName:
                return key, value

        return None, None

    @classmethod
    def __keyDescriptorByPythonName(cls, pythonName: str) -> Optional[KeyDescriptor]:
        if not pythonName in cls._keyDescriptors().keys():
            return None

        return cls._keyDescriptors()[pythonName]

    # - Encoding

    def __encodeKey(self, key: str) -> str:
        descriptor = self.__class__.__keyDescriptorByPythonName(key)

        if descriptor is None or descriptor.jsonName is None:
            return inflection.underscore(key)

        return descriptor.jsonName

    def _encodeValue(self, key: str, value: Any) -> Any:
        """
            Encodes python value into a json property
            Custom field serialization can be implemented by overriding this method

            Parameters
            ----------
            key : str
                python object variable name
            value : Any
                json object represented as an object from standard python library

            Returns
            -------
            Any -> encoded value of the object
        """

        descriptor = self.__class__.__keyDescriptorByPythonName(key)

        if descriptor is None or descriptor.pythonType is None:
            return value

        if issubclass(descriptor.pythonType, Enum):
            if descriptor.isList():
                return [element.value for element in value]

            return value.value

        if issubclass(descriptor.pythonType, UUID):
            if descriptor.isList():
                return [str(element) for element in value]

            return str(value)

        if issubclass(descriptor.pythonType, Codable):
            if descriptor.isList():
                return [descriptor.pythonType.encode(element) for element in value]

            return descriptor.pythonType.encode(value)

        if issubclass(descriptor.pythonType, datetime):
            if descriptor.isList():
                return [element.strftime(DATE_FORMAT) for element in value]

            return value.strftime(DATE_FORMAT)

        return value

    def encode(self) -> Dict[str, Any]:
        """
            Encodes python object into dictionary which contains
            only values representable by standard python library/types

            Returns
            -------
            Dict[str, Any] -> encoded object which can be serialized into json string
        """

        encodedObject: Dict[str, Any] = {}

        for key, value in self.__dict__.items():
            descriptor = self.__class__.__keyDescriptorByPythonName(key)

            # skip ignored fields for encoding
            if descriptor is not None and not descriptor.isEncodable:
                # print(f">> [Coretex] Skipping encoding for field: {key}")
                continue

            encodedKey = self.__encodeKey(key)
            encodedValue = self._encodeValue(key, value)

            encodedObject[encodedKey] = encodedValue

        return encodedObject

    # - Decoding

    @classmethod
    def __decodeKey(cls, key: str) -> str:
        descriptorKey, _ = cls.__keyDescriptorByJsonName(key)

        if descriptorKey is None:
            return inflection.camelize(key, False)

        return descriptorKey

    @classmethod
    def _decodeValue(cls, key: str, value: Any) -> Any:
        """
            Decodes a value of a single json field
            Custom logic can be implemented by overriding this method

            Parameters
            ----------
            key : str
                name of the json field
            value : Any
                value of the json field

            Returns
            -------
            Any -> decoded value of the json field
        """

        _, descriptor = cls.__keyDescriptorByJsonName(key)

        if descriptor is None or descriptor.pythonType is None:
            return value

        if issubclass(descriptor.pythonType, Enum):
            if descriptor.isList() and descriptor.collectionType is not None:
                return descriptor.collectionType([descriptor.pythonType(element) for element in value])

            return descriptor.pythonType(value)

        if issubclass(descriptor.pythonType, UUID):
            if descriptor.isList() and descriptor.collectionType is not None:
                return descriptor.collectionType([descriptor.pythonType(element) for element in value])

            return descriptor.pythonType(value)

        if issubclass(descriptor.pythonType, Codable):
            if descriptor.isList() and descriptor.collectionType is not None:
                return descriptor.collectionType([descriptor.pythonType.decode(element) for element in value])

            return descriptor.pythonType.decode(value)

        if issubclass(descriptor.pythonType, datetime):
            if descriptor.isList() and descriptor.collectionType is not None:
                return descriptor.collectionType([decodeDate(element) for element in value])

            return decodeDate(value)

        return value

    def _updateFields(self, encodedObject: Dict[str, Any]) -> None:
        """
            Updates the properties of object with new values

            Parameters
            ----------
            encodedObject : Dict[str, Any]
                json encoded object
        """

        for key, value in encodedObject.items():
            _, descriptor = self.__class__.__keyDescriptorByJsonName(key)

            # skip ignored fields for deserialization
            if descriptor is not None and not descriptor.isDecodable:
                # print(f">> [Coretex] Skipping decoding for field: {key}")
                continue

            decodedKey = self.__decodeKey(key)
            self.__dict__[decodedKey] = self._decodeValue(key, value)

    def onDecode(self) -> None:
        """
            Callback which is called once the object has been decoded
        """

        pass

    @classmethod
    def decode(cls, encodedObject: Dict[str, Any]) -> Self:
        """
            Decodes the json object into a python object

            Parameters
            ----------
            encodedObject : Dict[str, Any]
                json encoded object

            Returns
            -------
            Self -> Decoded python object
        """
        obj = cls()

        obj._updateFields(encodedObject)
        obj.onDecode()

        return obj



from pathlib import Path
from logging import FileHandler
from logging.handlers import WatchedFileHandler

import platform


class CoretexFileHandler(WatchedFileHandler):

    def reopenIfNeeded(self) -> None:
        # baseFilename is absoulte path to the log file
        path = Path(self.baseFilename)

        # If parent dir got deleted for some reason
        if not path.parent.exists():
            path.parent.mkdir(parents = True, exist_ok = True)

        super().reopenIfNeeded()


def createFileHandler(path: Path) -> FileHandler:
    """
        Creates a handler for writting logs to the file system

        Parameters
        ----------
        path : Path
            path of the log file

        Returns
        -------
        FileHandler -> if OS is Windows
        WatchedFileHandler -> if OS is Unix based
    """

    if platform.system() == "Windows":
        return FileHandler(path)

    return CoretexFileHandler(path)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Optional, Union
from typing_extensions import Self

from .metric_type import MetricType
from ....codable import Codable, KeyDescriptor


class Metric(Codable):

    name: str
    xLabel: str
    xType: str
    yLabel: str
    yType: str
    xRange: Optional[List[float]]
    yRange: Optional[List[float]]

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["name"] = KeyDescriptor("metric")

        return descriptors

    @classmethod
    def create(
        cls,
        name: str,
        xLabel: str,
        xType: Union[MetricType, int],
        yLabel: str,
        yType: Union[MetricType, int],
        xRange: Optional[List[float]] = None,
        yRange: Optional[List[float]] = None
    ) -> Self:

        """
            name : str
                name of Metric
            xLabel : str
                label of x axis which will be displayed
            xType : Union[MetricType, int]
                type of x axis which will be displayed
            yLabel : str
                label of y axis which will be displayed
            yType : Union[MetricType, int]
                type of y axis which will be displayed
            xRange : Optional[List[float]]
                range in which values will be displayed for x axis
            yRange : Optional[List[float]]
                range in which values will be displayed for y axis
        """

        if isinstance(xType, int):
            xType = MetricType(xType)

        if isinstance(yType, int):
            yType = MetricType(yType)

        obj = cls()

        obj.name = name
        obj.xLabel = xLabel
        obj.xType = xType.name
        obj.yLabel = yLabel
        obj.yType = yType.name
        obj.xRange = xRange
        obj.yRange = yRange

        return obj

    def extract(self) -> Optional[float]:
        return None



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any, Tuple
from typing_extensions import Self

import time
import json

from .severity import LogSeverity
from .utils import colorMessage
from ..utils import mathematicalRound


class Log:

    """
        Represents a single Coretex console log

        timestamp : float
            timestamp of the log
        message : str
            message of the log
        severity : LogSeverity
            severity of the log
    """

    def __init__(self, severity: LogSeverity, message: str) -> None:
        self.timestamp = mathematicalRound(time.time(), 6)
        self.severity = severity
        self.message = colorMessage(severity, message)

    def encode(self) -> Dict[str, Any]:
        return {
            "timestamp": self.timestamp,
            "severity": self.severity.value,
            "content": self.message,
            "type": 1  # Here for backwards compatibility
        }

    @classmethod
    def parse(cls, value: str) -> Tuple[Self, str]:
        try:
            jsonLog = json.loads(value)

            if not isinstance(jsonLog, dict):
                raise ValueError

            if len(jsonLog) != 2:
                raise ValueError

            return cls(LogSeverity(jsonLog["severity"]), jsonLog["message"].rstrip()), jsonLog["message"]
        except:
            return cls(LogSeverity.info, value.rstrip()), value



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type, Optional, List

from .metric import Metric, MetricType
from .predefined_metrics import *


def getClassForMetric(name: str) -> Optional[Type[Metric]]:
    if name == "disk_read":
        return MetricDiskRead

    if name == "disk_write":
        return MetricDiskWrite

    if name == "cpu_usage":
        return MetricCPUUsage

    if name == "ram_usage":
        return MetricRAMUsage

    if name == "swap_usage":
        return MetricSwapUsage

    if name == "gpu_usage":
        return MetricGPUUsage

    if name == "gpu_temperature":
        return MetricGPUTemperature

    if name == "upload_speed":
        return MetricUploadSpeed

    if name == "download_speed":
        return MetricDownloadSpeed

    return None


def createMetric(
    name: str,
    xLabel: str,
    xType: MetricType,
    yLabel: str,
    yType: MetricType,
    xRange: Optional[List[float]] = None,
    yRange: Optional[List[float]] = None
) -> Metric:

    metric = getClassForMetric(name)
    if metric is None:
        raise ValueError(f"[Coretex] Failed to create {name} metric")

    return metric.create(name, xLabel, xType, yLabel, yType, xRange, yRange)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime

from . import folder_manager
from .logging import initializeLogger, LogSeverity
from .configuration import CONFIG_DIR


def _initializeDefaultLogger() -> None:
    logName = datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f%z")
    logPath = folder_manager.coretexpylibLogs.joinpath(logName).with_suffix(".log")

    initializeLogger(LogSeverity.info, logPath, jsonOutput = False)


def _initializeCLILogger() -> None:
    logName = datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f%z")
    logPath = CONFIG_DIR / "logs"
    logPath.mkdir(exist_ok = True)

    initializeLogger(LogSeverity.info, logPath.joinpath(logName).with_suffix(".log"), jsonOutput = False)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple

from PIL import Image

import numpy as np


def resizeWithPadding(image: np.ndarray, width: int, height: int) -> Tuple[np.ndarray, int, int]:
    """
        Resizes the image while maintaining original aspect ratio,
        and filling the remaining space with symmetrical padding so
        that the original image is in the center

        Parameters
        ----------
        image : np.ndarray
            Input image as an array
        width : int
            Width of the output image
        height : int
            Height of the output image

        Returns
        -------
        Tuple[np.ndarray, int, int] -> Output image as numpy array,
        number of pixels of padding from top/bottom and number of pixels
        of padding from left/right
    """

    originalWidth = image.shape[1]
    originalHeight = image.shape[0]

    ratio = max(width, height) / max(originalWidth, originalHeight)

    newWidth = int(originalWidth * ratio)
    newHeight = int(originalHeight * ratio)

    resizedImage = Image.fromarray(image).resize((newWidth, newHeight))

    deltaW = width - newWidth
    deltaH = height - newHeight

    top = deltaH // 2
    left = deltaW // 2

    paddedImage = Image.new(resizedImage.mode, (width, height), 0)
    paddedImage.paste(resizedImage, (left, top))

    return np.array(paddedImage), top, left


def cropToWidth(image: np.ndarray) -> np.ndarray:
    """
        Crops the image to width while maintaining the center

        Parameters
        ----------
        image : np.ndarray
            Input image as an array

        Returns
        -------
        np.ndarray -> Output image that has been croped vertically
        to width
    """

    height, width = image.shape[:2]
    if height > width:
        startY = (height - width) // 2
        endY = startY + width
        image = image[startY:endY, 0:width]

    return image



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from enum import IntEnum

import logging

class LogSeverity(IntEnum):

    """
        List of all log severities supported by Coretex
    """

    fatal   = 1
    error   = 2
    warning = 3
    info    = 4
    debug   = 5

    @property
    def color(self) -> int:
        """
            ANSI color of the log severity

            Returns
            -------
            str -> color
        """

        if self == LogSeverity.fatal:
            return 31  # red

        if self == LogSeverity.error:
            return 31  # red

        if self == LogSeverity.warning:
            return 33  # yellow

        if self == LogSeverity.info:
            return 97  # white

        if self == LogSeverity.debug:
            return 33  # yellow

        raise RuntimeError(">> [Coretex] Invalid enum value")

    def getLevel(self) -> int:
        """
            Converts Coretex log severity into the
            equivalent log level from python std module logging

            Returns
            -------
            int -> python std module logging level

            Example
            -------
            >>> from coretex.logging import LogSeverity
            \b
            >>> print(LogSeverity.info.stdSeverity)
            20
        """

        if self == LogSeverity.fatal:
            return logging.FATAL

        if self == LogSeverity.error:
            return logging.ERROR

        if self == LogSeverity.warning:
            return logging.WARNING

        if self == LogSeverity.info:
            return logging.INFO

        if self == LogSeverity.debug:
            return logging.DEBUG

        raise RuntimeError(">> [Coretex] Invalid enum value")

    @property
    def prefix(self) -> str:
        """
            Prefix of the log severity

            Returns
            -------
            str -> prefix
        """

        return self.name.capitalize()

    @staticmethod
    def fromLevel(logLevel: int) -> LogSeverity:
        """
            Converts python std module logging level
            into the equivalent log severity from coretex

            Returns
            -------
            LogSeverity -> Coretex log severity

            Example
            -------
            >>> import logging
            \b
            >>> from coretex.logging.log_severity import LogSeverity
            >>> logSeverityValue = LogSeverity.fromStd(logging.INFO)
            >>> print(logSeverityValue)
            LogSeverity.info
        """

        if logLevel == logging.FATAL:
            return LogSeverity.fatal

        if logLevel == logging.ERROR:
            return LogSeverity.error

        if logLevel == logging.WARNING:
            return LogSeverity.warning

        if logLevel == logging.INFO:
            return LogSeverity.info

        if logLevel == logging.DEBUG:
            return LogSeverity.debug

        raise ValueError(">> [Coretex] Invalid enum value")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypeVar, Generic, Optional, List
from pathlib import Path

import json

from ...sample import Sample
from ...annotation import ImageDatasetClass, ImageDatasetClasses


SampleType = TypeVar("SampleType", bound = "Sample")


class BaseImageDataset(Generic[SampleType]):

    samples: List[SampleType]
    classes: ImageDatasetClasses
    path: Path

    @property
    def classesPath(self) -> Path:
        """
            Returns
            -------
            Path -> path to classes.json file
        """

        return self.path / "classes.json"

    def classByName(self, name: str) -> Optional[ImageDatasetClass]:
        for clazz in self.classes:
            if clazz.label == name:
                return clazz

        return None

    def _writeClassesToFile(self) -> None:
        self.path.mkdir(exist_ok = True)

        with open(self.classesPath, "w") as file:
            json.dump([clazz.encode() for clazz in self.classes], file)

    def saveClasses(self, classes: ImageDatasetClasses) -> bool:
        """
            Saves provided classes (including their color) to dataset.
            ImageDataset.classes property will be updated on successful save

            Parameters
            ----------
            classes : List[ImageDatasetClass]
                list of classes

            Returns
            -------
            bool -> True if dataset classes were saved, False if failed to save dataset classes

            Example
            -------
            >>> from coretex import ImageDatasetClass
            \b
            >>> labels = {"car", "bicycle", "person", "tree"}
            >>> imgDatasetClasses = ImageDatasetClass.generate(labels)
            >>> imageDatasetObj.saveClasses(imgDatasetClasses)
            True
        """

        self.classes = classes
        self._writeClassesToFile()

        return True



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from importlib.metadata import version as getLibraryVersion

import click

from .commands.login import login
from .commands.model import model
from .commands.node import node
from .commands.task import run
from .commands.project import project

from .modules import ui
from .modules.intercept import ClickExceptionInterceptor
from .modules.utils import updateLib

from ..utils.process import CommandException


@click.command()
def version() -> None:
    version = getLibraryVersion("coretex")
    ui.stdEcho(f"Coretex {version}")


@click.command()
def update() -> None:
    try:
        ui.progressEcho("Updating coretex...")
        updateLib()
        ui.successEcho("Coretex successfully updated!")
    except CommandException:
        ui.errorEcho("Failed to update coretex.")


@click.group(cls = ClickExceptionInterceptor)
def cli() -> None:
    pass

cli.add_command(login)
cli.add_command(model)
cli.add_command(project)
cli.add_command(node)
cli.add_command(run)
cli.add_command(version)
cli.add_command(update)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict

from .image_sample_data import AnnotatedImageSampleData
from .local_image_sample import LocalImageSample
from ..network_sample import NetworkSample
from ...annotation import CoretexImageAnnotation
from ....networking import networkManager, NetworkRequestError


class ImageSample(NetworkSample[AnnotatedImageSampleData], LocalImageSample):

    """
        Represents the generic image sample\n
        Contains basic properties and functionality for all image sample classes\n
        The class has several methods that allow users to access and
        manipulate image data and annotations, as well as to create new image samples
    """

    def __init__(self) -> None:
        NetworkSample.__init__(self)

    def saveAnnotation(self, coretexAnnotation: CoretexImageAnnotation) -> bool:
        # Encrypted sample must be downloaded for annotation to be updated
        if self.isEncrypted:
            self.download(decrypt = True)

        # Only save annotation locally if it is downloaded
        if self.zipPath.exists():
            self.unzip()

            super().saveAnnotation(coretexAnnotation)

        if self.isEncrypted:
            try:
                self._overwriteSample(self.zipPath)
                return True
            except NetworkRequestError:
                return False
        else:
            parameters = {
                "id": self.id,
                "data": coretexAnnotation.encode()
            }

            response = networkManager.post("session/save-annotations", parameters)
            return not response.hasFailed()

    def saveMetadata(self, metadata: Dict[str, Any]) -> None:
        # Encrypted sample must be downloaded for metadata to be updated
        if self.isEncrypted:
            self.download(decrypt = True)

        # Only save metadata locally if it is downloaded
        if self.zipPath.exists():
            self.unzip()

            super().saveMetadata(metadata)

        if self.isEncrypted:
            self._overwriteSample(self.zipPath)
        else:
            parameters = {
                "id": self.id,
                "data": metadata
            }

            response = networkManager.post("session/save-metadata", parameters)
            if response.hasFailed():
                raise NetworkRequestError(response, f"Failed to upload metadata for sample \"{self.name}\"")



from typing import Optional, Dict, Any

import logging

import click

from . import ui
from ...entities import Project, ProjectType, ProjectVisibility
from ...networking import NetworkRequestError
from ...configuration import loadConfig, saveConfig


def selectProject(projectId: int) -> None:
    config = loadConfig()
    config["projectId"] = projectId
    saveConfig(config)


def selectProjectType() -> ProjectType:
    availableProjectTypes = {
        "Computer Vision": ProjectType.computerVision,
        "Motion Recognition": ProjectType.motionRecognition,
        "Bioinformatics": ProjectType.bioInformatics,
        "Other": ProjectType.other
    }

    choices = list(availableProjectTypes.keys())
    selectedChoice = ui.arrowPrompt(choices, "Specify type of project that you wish to create")

    selectedProjectType = availableProjectTypes[selectedChoice]
    ui.stdEcho(f"You've chosen: {selectedChoice}")
    return selectedProjectType


def selectProjectVisibility() -> ProjectVisibility:
    availableProjectVisibilities = {
        "Private": ProjectVisibility.private,
        "Public": ProjectVisibility.public,
    }

    choices = list(availableProjectVisibilities.keys())
    selectedChoice = ui.arrowPrompt(choices, "Specify visibility of project:")

    selectedProjectVisibility = availableProjectVisibilities[selectedChoice]
    ui.stdEcho(f"You've chosen: {selectedChoice}")
    return selectedProjectVisibility


def promptProjectCreate(message: str, name: str) -> Optional[Project]:
    if not click.confirm(message, default = True):
        return None

    selectedProjectType = selectProjectType()

    try:
        project = Project.createProject(name, selectedProjectType)
        ui.successEcho(f"Project \"{name}\" created successfully.")
        ui.stdEcho(f"A new Project has been created. You can open it by clicking on this URL {ui.outputUrl(project.entityUrl())}.")
        return project
    except NetworkRequestError as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise click.ClickException(f"Failed to create project \"{name}\".")


def promptProjectSelect() -> Optional[Project]:
    name = ui.clickPrompt("Specify project name that you wish to select")

    ui.progressEcho("Validating project...")
    try:
        project = Project.fetchOne(name = name)
        ui.successEcho(f"Project \"{project}\" selected successfully!")
        selectProject(project.id)
    except ValueError:
        ui.errorEcho(f"Project \"{name}\" not found.")
        newProject = promptProjectCreate("Do you want to create a project with that name?", name)
        if newProject is None:
            return None

        selectProject(project.id)

    return project


def createProject(name: Optional[str] = None, projectType: Optional[int] = None, description: Optional[str] = None) -> Project:
    if name is None:
        name = ui.clickPrompt("Please enter name of the project you want to create", type = str)

    if projectType is None:
        projectType = selectProjectType()
    else:
        projectType = ProjectType(projectType)

    if description is None:
        description = ui.clickPrompt("Please enter your project's description", type = str, default = "", show_default = False)

    try:
        project = Project.createProject(name, projectType, description = description)
        ui.successEcho(f"Project \"{name}\" created successfully.")
        ui.stdEcho(f"A new Project has been created. You can open it by clicking on this URL {ui.outputUrl(project.entityUrl())}.")
        return project
    except NetworkRequestError as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise click.ClickException(f"Failed to create project \"{name}\".")


def getProject(name: Optional[str], config: Dict[str, Any]) -> Optional[Project]:
    projectId = config.get("projectId")
    if name is not None:
        try:
            return Project.fetchOne(name = name)
        except:
            if projectId is None:
                return promptProjectCreate("Project not found. Do you want to create a new Project with that name?", name)

            return Project.fetchById(projectId)

    if projectId is None:
        ui.stdEcho("To use this command you need to have a Project selected.")
        if click.confirm("Would you like to select an existing Project?", default = True):
            return promptProjectSelect()

        if not click.confirm("Would you like to create a new Project?", default = True):
            return None

        return createProject(name)

    return Project.fetchById(projectId)


def isProjectSelected() -> bool:
    config = loadConfig()

    if config.get("projectId") is None:
        return False

    try:
        Project.fetchById(config["projectId"])
        return True
    except NetworkRequestError:
        return False



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, List, Dict, Any
from typing_extensions import Self
from pathlib import Path

import yaml

from ...entities import BaseParameter, parameter_factory
from ...codable import Codable, KeyDescriptor


TASK_CONFIG_PATH = Path(".", "task.yaml")


class ParamGroup(Codable):

    name: str
    params: Optional[List[BaseParameter]]

    @classmethod
    def _decodeValue(cls, key: str, value: Any) -> Any:
        if key == "params":
            return [parameter_factory.create(obj) for obj in value]

        return super()._decodeValue(key, value)


class TaskConfig(Codable):

    paramGroups: Optional[List[ParamGroup]]

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["paramGroups"] = KeyDescriptor("param_groups", ParamGroup, list)

        return descriptors

    @classmethod
    def decode(cls, params: dict) -> Self:
        if params.get("param_groups") is None:
            params["param_groups"] = []

        return super().decode(params)


def readTaskConfig() -> List[BaseParameter]:
    parameters: List[BaseParameter] = []

    if not TASK_CONFIG_PATH.exists():
        return []

    with TASK_CONFIG_PATH.open("r") as configFile:
        config = TaskConfig.decode(yaml.safe_load(configFile))

        if config.paramGroups is not None:
            for group in config.paramGroups:
                if group.params is not None:
                    parameters.extend(group.params)

    return parameters



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Optional, List, Dict, Set
from uuid import UUID

import uuid
import random

from ....codable import Codable, KeyDescriptor


class ImageDatasetClass(Codable):

    """
        Image Dataset class metadata

        Properties
        ----------
        classIds : List[UUID]
            list of all uuids connected to this class
        label : str
            name of the class
        color : str
            color of the class
    """

    classIds: List[UUID]
    label: str
    color: str

    def __init__(self, label: Optional[str] = None, color: Optional[str] = None):
        if label is None:
            label = ""

        if color is None:
            color = ""

        self.classIds = [uuid.uuid4()]
        self.label = label
        self.color = color

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["classIds"] = KeyDescriptor("ids", UUID, list)
        descriptors["label"] = KeyDescriptor("name")

        return descriptors

    @classmethod
    def generate(cls, labels: Set[str]) -> ImageDatasetClasses:
        """
            Generates list of Coretex classes based on provided
            labels (class names)

            Parameters
            ----------
            labels : Set[str]
                list of labels (class names)

            Returns
            -------
            ImageDatasetClasses -> list of ImageDatasetClass (Coretex class metadata) objects

            Example
            -------
            >>> from coretex import ImageDatasetClass
            \b
            >>> labels = {"car", "bicycle"}
            >>> imgDatasetClasses = ImageDatasetClass.generate(labels)
            >>> print(classes)
            [
                ImageDatasetClass(
                    classIds = [UUID("81add001-1c9c-4949-8b73-3599f1d0de9d")],
                    name = "car",
                    color = "#bcb86b"
                ),
                ImageDatasetClass(
                    classIds = [UUID("d710019b-f28f-40ab-aa65-e13df949beff")],
                    name = "bicycle",
                    color = "#cbc66b"
                )
            ]
        """

        colors: Set[str] = set()

        while len(colors) != len(labels):
            color = f'#{"%06x" % random.randint(0, 0xFFFFFF)}'
            colors.add(color)

        return ImageDatasetClasses(
            [cls(label, color) for label, color in zip(labels, colors)]
        )


class ImageDatasetClasses(List[ImageDatasetClass]):

    """
        List of Image Dataset class metadata

        Properties
        ----------
        labels : List[str]
            list of the classes names
    """

    @property
    def labels(self) -> List[str]:
        labels = [element.label for element in self]
        labels.sort()

        return labels

    def classById(self, classId: UUID) -> Optional[ImageDatasetClass]:
        """
            Retrieves a Image dataset class based on provided ID

            Parameters
            ----------
            classID : UUID
                id of class

            Returns
            -------
            Optional[ImageDatasetClasses] -> fetched class if provided ID
            is found in list of class IDs, None otherwise

            Examples
            --------
            >>> from coretex import ImageDataset
            \b
            >>> dataset = ImageDataset.fetchById(1023)
            >>> classObj = dataset.classes.classById(UUID("d710019b-f28f-40ab-aa65-e13df949beff"))
            \b
            >>> if classObj is not None:
                    print(classObj.classIds)
                    print(classObj.label)
                    print(classObj.color)
            UUID("d710019b-f28f-40ab-aa65-e13df949beff")
            "bicycle"
            "#d06df5"
        """

        for element in self:
            for other in element.classIds:
                if str(classId) == str(other):
                    return element

        return None

    def classByLabel(self, label: str) -> Optional[ImageDatasetClass]:
        """
            Retrieves a Image dataset class based on provided label

            Parameters
            ----------
            label : str
                name of class

            Returns
            -------
            Optional[ImageDatasetClasses] -> fetched class if provided label
            is found in list of class labels, None otherwise
        """

        for element in self:
            if element.label == label:
                return element

        return None

    def labelIdForClassId(self, classId: UUID) -> Optional[int]:
        """
            Retrieves a label ID based on provided class ID

            Parameters
            ----------
            classId : UUID
                id of class

            Returns
            -------
            Optional[int] -> label ID if provided class ID
            is found in list of class IDs, None otherwise

            Example
            -------
            >>> from coretex import ImageDataset
            \b
            >>> dataset = ImageDataset.fetchById(1023)
            >>> labelId = dataset.classes.labelIdForClassId(UUID("d710019b-f28f-40ab-aa65-e13df949beff"))
            >>> print(labelId)
            1

        """

        clazz = self.classById(classId)
        if clazz is None:
            return None

        try:
            return self.labels.index(clazz.label)
        except ValueError:
            return None

    def labelIdForClass(self, clazz: ImageDatasetClass) -> Optional[int]:
        """
            Retrieves a label ID based on provided ImageDatasetClass object

            Parameters
            ----------
            clazz : ImageDatasetClass
                Image Dataset class metadata object

            Returns
            -------
            Optional[int] -> label ID if provided class exists
            in a list of ImageDataset classes otherwise None

            Example
            -------
            >>> from coretex import ImageDatasetClass
            \b
            >>> dataset = ImageDataset.fetchById(1023)
            >>> labelId = dataset.classes.labelIdForClass(dataset.classes[0])
            >>> print(labelId)
            0
        """

        return self.labelIdForClassId(clazz.classIds[0])

    def exclude(self, excludedClasses: List[str]) -> None:
        """
            Excludes classes that are provided in excludedClasses list

            Parameters
            ----------
            excludedClasses : List[str]
                list of classes that will be excluded
        """

        classes = [
            element for element in self
            if element.label not in excludedClasses
        ]

        self.clear()
        self.extend(classes)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Dict, List, Union, Any
from typing_extensions import Self
from enum import IntEnum
from pathlib import Path

from ... import folder_manager
from ...codable import Codable, KeyDescriptor
from ...networking import networkManager, FileData
from ...utils import guessMimeType


class ArtifactType(IntEnum):

    """
        Available types of Artifacts on Coretex
    """

    directory = 1
    file      = 2


class Artifact(Codable):

    """
        Artifact class represents a single result of a run\n
        Result can be file of any kind

        Properties
        ----------
        artifactType : ArtifactType
            type of Artifact
        remoteFilePath : str
            path of Artifact on Coretex
        size : Optional[int]
            size of Artifact in bytes (not required)
        mimeType : str
            mimeType of Artifact
        timestamp : int
            current timestamp
        taskRunId : int
            id of run
    """

    artifactType: ArtifactType
    remoteFilePath: str
    size: Optional[int]
    mimeType: str
    timestamp: int
    taskRunId: int

    @property
    def localFilePath(self) -> Path:
        """
            Represents the local path of the Artifact

            Returns
            -------
            Path -> local path to Artifact
        """

        return folder_manager.getArtifactsFolder(self.taskRunId) / self.remoteFilePath

    @property
    def isDirectory(self) -> bool:
        """
            Returns
            -------
            bool -> True if Artifact type is directory
        """

        return self.artifactType == ArtifactType.directory

    @property
    def isFile(self) -> bool:
        """
            Returns
            -------
            bool -> True if Artifact type is file
        """

        return self.artifactType == ArtifactType.file

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["artifactType"] = KeyDescriptor("type", ArtifactType)
        descriptors["remoteFilePath"] = KeyDescriptor("path")
        descriptors["timestamp"] = KeyDescriptor("ts")
        descriptors["taskRunId"] = KeyDescriptor(isEncodable = False)

        return descriptors

    @classmethod
    def create(
        cls,
        taskRunId: int,
        localFilePath: Union[Path, str],
        remoteFilePath: str,
        mimeType: Optional[str] = None
    ) -> Optional[Self]:

        if mimeType is None:
            # If guessing fails, fallback to binary type
            try:
                mimeType = guessMimeType(localFilePath)
            except:
                mimeType = "application/octet-stream"

        parameters = {
            "model_queue_id": taskRunId,
            "path": remoteFilePath
        }

        files = [
            FileData.createFromPath("file", localFilePath, mimeType = mimeType)
        ]

        response = networkManager.formData("artifact/upload-file", parameters, files)
        if response.hasFailed():
            return None

        artifact = cls.decode(response.getJson(dict))
        artifact.taskRunId = taskRunId

        return artifact

    def download(self) -> bool:
        """
            Downloads Artifact from Coretex.ai

            Returns
            -------
            bool -> False if response has failed, True otherwise
        """

        params = {
            "model_queue_id": self.taskRunId,
            "path": self.remoteFilePath
        }

        return not networkManager.download("artifact/download-file", str(self.localFilePath), params).hasFailed()

    @classmethod
    def fetchAll(cls, taskRunId: int, path: Optional[str] = None, recursive: bool = False) -> List[Self]:
        """
            Fetch all Artifacts from Coretex.ai for the specified run

            Parameters
            ----------
            taskRunId : int
                id of run
            path : Optional[str]
                local path where u want to store fetched Artifacts
            recursive : bool
                True if you want list to be sorted recursively, False otherwise
        """

        params: Dict[str, Any] = {
            "model_queue_id": taskRunId,
        }

        if path is not None:
            params["path"] = path

        response = networkManager.get("artifact/list-contents", params)
        if response.hasFailed():
            return []

        artifacts = [cls.decode(element) for element in response.getJson(list)]

        for artifact in artifacts:
            artifact.taskRunId = taskRunId

            if recursive and artifact.isDirectory:
                artifacts.extend(
                    cls.fetchAll(taskRunId, artifact.remoteFilePath)
                )

        return artifacts



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any

import inflection

from .utils import getValueParamType
from .base_parameter import BaseParameter
from .parameter_type import ParameterType
from .parameters import *
from ...secret import AWSSecret, GitSecret, CredentialsSecret


def cleanParamDict(value: Dict[str, Any]) -> Dict[str, Any]:
    if not "data_type" in value:
        value["data_type"] = None

    if not "value" in value:
        value["value"] = None

    if not "required" in value:
        value["required"] = False

    if not "description" in value:
        value["description"] = ""

    dataName: str = value["name"]
    dataType = value["data_type"]
    dataValue = value["value"]

    if dataType is None and dataValue is None:
        value["data_type"] = ParameterType.string.value

    if dataType is None and dataValue is not None:
        value["data_type"] = getValueParamType(dataValue, dataName).value

    return value

def create(value: Dict[str, Any]) -> BaseParameter:
    cleanParamDict(value)
    dataType = value.get("data_type")

    if not isinstance(dataType, str):
        raise ValueError("\"data_type\" is not of type \"str\"")

    # Convert snake_case keys into camelCase
    value = {
        inflection.camelize(key, uppercase_first_letter = False): value
        for key, value in value.items()
    }

    parameterType = ParameterType(dataType)

    if parameterType == ParameterType.integer:
        return IntParameter(**value)

    if parameterType == ParameterType.floatingPoint:
        return FloatParameter(**value)

    if parameterType == ParameterType.string:
        return StrParameter(**value)

    if parameterType == ParameterType.boolean:
        return BoolParameter(**value)

    if parameterType == ParameterType.dataset:
        return DatasetParameter(**value)

    if parameterType == ParameterType.model:
        return ModelParameter(**value)

    if parameterType == ParameterType.imuVectors:
        return IMUVectorsParameter(**value)

    if parameterType == ParameterType.enum:
        return EnumParameter(**value)

    if parameterType == ParameterType.range:
        return RangeParameter(**value)

    if parameterType == ParameterType.awsSecret:
        return SecretParameter(AWSSecret, **value)

    if parameterType == ParameterType.gitSecret:
        return SecretParameter(GitSecret, **value)

    if parameterType == ParameterType.credentialsSecret:
        return SecretParameter(CredentialsSecret, **value)

    if parameterType == ParameterType.intList:
        return ListIntParameter(**value)

    if parameterType == ParameterType.floatList:
        return ListFloatParameter(**value)

    if parameterType == ParameterType.strList:
        return ListStrParameter(**value)

    if parameterType == ParameterType.datasetList:
        return ListDatasetParameter(**value)

    if parameterType == ParameterType.modelList:
        return ListModelParameter(**value)

    if parameterType == ParameterType.enumList:
        return ListEnumParameter(**value)

    if parameterType == ParameterType.awsSecretList:
        return ListSecretParameter(AWSSecret, **value)

    if parameterType == ParameterType.gitSecretList:
        return ListSecretParameter(GitSecret, **value)

    if parameterType == ParameterType.credentialsSecretList:
        return ListSecretParameter(CredentialsSecret, **value)

    raise ValueError(f"Unknown parameter type \"{parameterType}\"")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Dict, Optional, Tuple

import logging

from ..base_parameter import BaseParameter
from ..utils import validateRangeStructure
from ....project import ProjectType


RANGE_LIMIT = 2^32


class RangeParameter(BaseParameter[Dict[str, int]]):

    @property
    def types(self) -> List[type]:
        return NotImplemented

    def validate(self) -> Tuple[bool, Optional[str]]:
        if not isinstance(self.value, dict):
            return False, None

        isValid, message = validateRangeStructure(self.name, self.value, self.required)
        if not isValid:
            return isValid, message

        if self.value["to"] > RANGE_LIMIT:
            return False, f"Range value \"to\" must not exceed 2^32"

        if not self.value["to"] > self.value["from"]:
            return False, f"Range value \"to\" must not be greater then \"from\""

        if not self.value["step"] <= self.value["to"] - self.value["from"]:
            return False, f"Range value \"step\" must be lower or equal to the distance between \"from\" and \"to\""

        return True, None

    def parseValue(self, type_: ProjectType) -> Optional[Any]:
        if self.value is None:
            return None

        return range(self.value["from"], self.value["to"], self.value["step"])

    def overrideValue(self, value: Optional[Any]) -> Optional[Any]:
        if value is None or self.value is None:
            return None

        try:
            args = value.split(" ")
            parsedValue: Dict[str, int] = {}

            # In case the only 2 values are entered, they will be treaded as "from" and "to"
            # Default step is 1
            if len(args) not in [2, 3]:
                return None

            parsedValue["from"] = int(args[0])
            parsedValue["to"] = int(args[1])
            parsedValue["step"] = int(args[2]) if len(args) == 3 else 1

            return parsedValue
        except ValueError as e:
            logging.getLogger("coretexpylib").warning(f">> [Coretex] Failed to override range parameter \"{self.name}\". | {e}")
            return self.value



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

import click

from ..modules import ui, project_utils, utils, user
from ...entities import Project, ProjectVisibility
from ...networking import RequestFailedError
from ...configuration import loadConfig


@click.command()
@click.option("--name", "-n", type = str, help = "Project name")
@click.option("--type", "-t", "projectType", type = int, help = "Project type")
@click.option("--description", "-d", type = str, help = "Project description")
def create(name: Optional[str], projectType: Optional[int], description: Optional[str]) -> None:
    project = project_utils.createProject(name, projectType, description)

    selectNewProject = ui.clickPrompt("Do you want to select the new project as default? (Y/n)", type = bool, default = True)
    if selectNewProject:
        project_utils.selectProject(project.id)
        ui.successEcho(f"Project \"{project.name}\" successfully selected.")


@click.command()
@click.option("--project", "-p", type = str, help = "Project name")
@click.option("--name", "-n", type = str, help = "New Project name")
@click.option("--description", "-d", type = str, help = "New Project description")
def edit(project: Optional[str], name: Optional[str], description: Optional[str]) -> None:
    config = loadConfig()
    defaultProjectId = config.get("projectId")
    if defaultProjectId is None and project is None:
        ui.errorEcho(f"To use edit command you need to specifiy project name using \"--project\" or \"-p\" flag, or you can select default project using \"coretex project select\" command.")
        return

    if project is None and defaultProjectId is not None:
        selectedProject = Project.fetchById(defaultProjectId)
    else:
        selectedProject = Project.fetchOne(name = project)

    if name is None:
        name = ui.clickPrompt("Please enter new name for your project", type = str, default = selectedProject.name)

    if description is None:
        description = ui.clickPrompt("Please enter new description for your project", type = str, default = selectedProject.description, show_default = False)

    try:
        selectedProject.update(name = selectedProject.name, description = description)
        selectedProject.updateVisibility(ProjectVisibility.private)
        ui.successEcho(f"Project id \"{selectedProject.id}\" successfully edited.")

    except RequestFailedError:
        raise click.ClickException(f"Failed to edit project \"{selectedProject.name}\".")


@click.command()
@click.argument("name", type = str)
def select(name: str) -> None:
    project: Optional[Project] = None

    ui.progressEcho("Validating project...")

    try:
        project = Project.fetchOne(name = name)
        ui.successEcho(f"Project \"{name}\" selected successfully!")
        project_utils.selectProject(project.id)
    except ValueError:
        ui.errorEcho(f"Project \"{name}\" not found.")
        project = project_utils.promptProjectCreate("Do you want to create a project with that name?", name)
        if project is None:
            return

        project_utils.selectProject(project.id)


@click.group()
@utils.onBeforeCommandExecute(user.initializeUserSession)
def project() -> None:
    pass


project.add_command(create, "create")
project.add_command(edit, "edit")
project.add_command(select, "select")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

class DataBuffer:

    def __init__(self) -> None:
        self.data = bytearray()
        self.position = 0

    @property
    def remaining(self) -> int:
        return len(self.data) - self.position

    def append(self, data: bytes) -> None:
        """
            Appends new data to the end of the buffer

            Parameters
            ----------
            data : bytes
                data which will be appended
        """

        self.data.extend(data)

    def get(self) -> int:
        """
            Reads one byte from the buffer and increments
            its internal state by 1

            Returns
            -------
            int -> byte which was read

            Raises
            ------
            OverflowError -> if there are no more bytes to read
        """

        if len(self.data) <= self.position:
            raise OverflowError("All data was extracted from buffer")

        value = self.data[self.position]
        self.position += 1

        return value

    def getBytes(self, count: int) -> bytes:
        """
            Reads N number of bytes from the buffer and increments
            its internal state by N

            Parameters
            ----------
            count : int
                number of bytes to read

            Returns
            -------
            bytes -> bytes which were read

            Raises
            ------
            OverflowError -> if count exceeds number of available bytes
        """

        if len(self.data) < (self.position + count):
            raise OverflowError("Tried to extract more than than what the buffer has")

        values = self.data[self.position:self.position + count]
        self.position += count

        return values

    def getRemaining(self) -> bytes:
        """
            Reads remaining bytes from the buffer

            Returns
            -------
            bytes -> bytes which were read
        """

        remaining = len(self.data) - self.position
        return self.getBytes(remaining)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type, TypeVar, Optional, Iterator, Dict, Any
from contextlib import contextmanager

from .network_dataset import NetworkDataset


T = TypeVar("T", bound = NetworkDataset)


@contextmanager
def createDataset(
    type_: Type[T],
    name: str,
    projectId: int,
    meta: Optional[Dict[str, Any]] = None
) -> Iterator[T]:

    """
        Creates a new dataset with the provided name and type
        and finalizes its state in data base

        Parameters
        ----------
        type_ : Type[T]
            type of dataset which will be created
        name : str
            dataset name
        projectId : int
            project for which the dataset will be created
        metadataPath : Union[Path, str]
            path the zipped metadata file

        Returns
        -------
        The created sequence dataset object or None if creation failed

        Raises
        -------
        NetworkRequestError -> If dataset creation failed

        Example
        -------
        >>> from coretex import createDataset
        \b
        >>> with createDataset("dummyDataset", 123, pathToMetadata) as dataset:
        >>>     print(f"Dataset with id \"{dataset.id}\" and name \"{dataset.name}\" created")
    """

    dataset = type_.create(
        name = name,
        project_id = projectId,
        meta = meta
    )

    yield dataset
    if not dataset.finalize():
        raise ValueError(">> [Coretex] Failed to finalize dataset")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple
from base64 import b64decode

import os

from Crypto.PublicKey import RSA
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding

from .random_generator import Random


def _generateRsaNumbers(bits: int, seed: bytes) -> Tuple[int, int, int, int, int]:
    """
        Randomly generates numbers used for calculating RSA public
        and private keys

        Parameters
        ----------
        bits : int
            # of bits in the "n" public number
        seeed : bytes
            Array of bytes which represent seed for generating key in a deterministic way.
            The entropy of the generated key is equal to the entropy of the seed.

        Returns
        Tuple[int, int, int, int, int] -> p, q, n, e, d RSA values, e is always equal to 65537
    """

    random = Random(seed, hashes.SHA256())
    numbers = RSA.generate(bits, randfunc = random.getRandomBytes, e = 65537)

    return numbers.p, numbers.q, numbers.n, numbers.e, numbers.d


def generateKey(length: int, seed: bytes) -> rsa.RSAPrivateKey:
    """
        Generates RSA key-pair with the provided length and seed

        Parameters
        ----------
        length : int
            length of the key-pair
        seed : bytes
            Array of bytes which represent seed for generating key in a deterministic way.
            The entropy of the generated key is equal to the entropy of the seed.

        Returns
        -------
        RSAPrivateKey -> cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey object
    """

    # p -> large random prime number
    # q -> large random prime number
    # n -> p * q
    # e -> 65537
    # d -> inverse(e, lcm(p - 1, q - 1))
    p, q, n, e, d = _generateRsaNumbers(length, seed)

    dmp1 = rsa.rsa_crt_dmp1(d, p)
    dmq1 = rsa.rsa_crt_dmq1(d, q)
    iqmp = rsa.rsa_crt_iqmp(p, q)

    privateNumbers = rsa.RSAPrivateNumbers(p, q, d, dmp1, dmq1, iqmp, rsa.RSAPublicNumbers(e, n))
    return privateNumbers.private_key()


def getPrivateKeyBytes(key: rsa.RSAPrivateKey) -> bytes:
    """
        Converts the provided key into a byte array.
        Encoding is PEM, Format is PKCS8

        Parameters
        ----------
        key : RSAPrivateKey
            cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey object

        Returns
        -------
        bytes -> RSA private key serialized to bytes
    """

    return key.private_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PrivateFormat.PKCS8,
        encryption_algorithm = serialization.NoEncryption()
    )


def getPublicKeyBytes(key: rsa.RSAPublicKey) -> bytes:
    """
        Converts the provided key into a byte array.
        Encoding is PEM, Format is PKCS1

        Parameters
        ----------
        key : RSAPublicKey
            cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey object

        Returns
        -------
        bytes -> RSA public key serialized to bytes
    """

    return key.public_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PublicFormat.PKCS1
    )


def privateKeyFromBytes(keyBytes: bytes) -> rsa.RSAPrivateKey:
    """
        Creates private key object from provided private
        key bytes.

        Parameters
        ----------
        keyBytes : bytes
            private key bytes

        Returns
        -------
        rsa.RSAPrivateKey -> private key object
    """

    key = serialization.load_pem_private_key(keyBytes, password = None)

    if not isinstance(key, rsa.RSAPrivateKey):
        raise TypeError

    return key


def getPrivateKey() -> rsa.RSAPrivateKey:
    """
        Retrieves RSA private key stored in CTX_PRIVATE_KEY
        environment variable in base64 format.

        Returns
        -------
        rsa.RSAPrivateKey -> private key object
    """

    if "CTX_PRIVATE_KEY" not in os.environ:
        raise RuntimeError(f"\"CTX_PRIVATE_KEY\" environment variable not set")

    privateKey = b64decode(os.environ["CTX_PRIVATE_KEY"])
    return privateKeyFromBytes(privateKey)


def decrypt(key: rsa.RSAPrivateKey, value: bytes) -> bytes:
    """
        Decrypts ciphertext encrypted using provided key-pair.
        It is assumed that PKCS#1 padding was used during encryption.

        Parameters
        ----------
        key : rsa.RSAPrivateKey
            private key object
        value : bytes
            ciphertext

        Returns
        -------
        bytes -> decrypted ciphertext
    """

    return key.decrypt(value, padding.PKCS1v15())



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Set

import os
import glob
import xml.etree.ElementTree as ET

from .shared import getTag, toInt
from .instance_extractor import InstanceExtractor
from ...base_converter import BaseConverter
from .....entities import CoretexImageAnnotation


class PascalSegConverter(BaseConverter):

    """
        Represents the Converter from Pascal VOC 2012 Format to Cortex Format
    """

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__imagesPath = os.path.join(datasetPath, "JPEGImages")
        self.__segmentationPath = os.path.join(datasetPath, "SegmentationObject")

        annotations = os.path.join(datasetPath, "Annotations")
        self.__fileNames = glob.glob(os.path.join(annotations, "*.xml"))

    def _dataSource(self) -> List[str]:
        return self.__fileNames

    def _extractLabels(self) -> Set[str]:
        labels: Set[str] = set()

        for filename in self.__fileNames:
            tree = ET.parse(filename)
            root = tree.getroot()
            objects = root.findall("object")

            for obj in objects:
                labelElement = obj.find("name")
                if labelElement is None:
                    continue

                label = labelElement.text
                if label is None:
                    continue

                labels.add(label)

        return labels

    def __extractImageAnnotation(self, root: ET.Element) -> None:
        fileName = getTag(root, "filename")
        if fileName is None:
            return

        baseFileName = os.path.splitext(fileName)[0]
        filenamePNG = f"{baseFileName}.png"

        if not os.path.exists(os.path.join(self.__imagesPath, fileName)):
            return

        instanceExtractor = InstanceExtractor(self._dataset)
        instances = instanceExtractor.extractInstances(root, filenamePNG, self.__segmentationPath)

        size = root.find('size')
        if size is None:
            return

        width, height = toInt(size, "width", "height")
        if width is None or height is None:
            return

        coretexAnnotation = CoretexImageAnnotation.create(fileName, width, height, instances)
        self._saveImageAnnotationPair(os.path.join(self.__imagesPath, fileName), coretexAnnotation)

    def _extractSingleAnnotation(self, fileName: str) -> None:
        tree = ET.parse(fileName)
        root = tree.getroot()

        self.__extractImageAnnotation(root)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple

import os
import logging
import shutil

from py3nvml import py3nvml

import psutil


def getCpuUsage() -> float:
    """
        Returns
        -------
        float -> CPU usage as percentage since last call of this function
    """

    return psutil.cpu_percent()


def getRamUsage() -> float:
    """
        Returns
        -------
        float -> Used RAM memory as percentage
    """

    return psutil.virtual_memory().percent


def getAvailableRam() -> int:
    """
        Returns
        -------
        int -> total RAM memory in GB as int
    """

    memory = psutil.virtual_memory()
    return int(memory.total / (1024 ** 3))


def getAvailableCpuCount() -> int:
    availableCPUs = os.cpu_count()
    return availableCPUs if availableCPUs is not None else 1


def getGpuUsage() -> float:
    """
        py3nvml init must be called before calling this function
        otherwise it will raise an exception

        Returns
        -------
        float -> GPU usage as percentage
    """

    handle = py3nvml.nvmlDeviceGetHandleByIndex(0)
    utilization = py3nvml.nvmlDeviceGetUtilizationRates(handle)

    if isinstance(utilization, py3nvml.c_nvmlUtilization_t):
        return float(utilization.gpu)

    logging.getLogger("coretexpylib").debug(">> [Coretex] Failed to extract gpu usage metric")
    return 0


def getGpuTemperature() -> float:
    """
        py3nvml init must be called before calling this function
        otherwise it will raise an exception

        Returns
        -------
        float -> GPU temperature
    """

    handle = py3nvml.nvmlDeviceGetHandleByIndex(0)
    temperature = py3nvml.nvmlDeviceGetTemperature(handle, py3nvml.NVML_TEMPERATURE_GPU)

    return float(temperature)


def getSwapUsage() -> float:
    """
        Returns
        -------
        float -> Used swap memory as percentage
    """

    return psutil.swap_memory().percent


def getDiskRead() -> float:
    """
        Returns
        -------
        float -> total amount of bytes read from disk
    """

    counters = psutil.disk_io_counters()

    if counters is None:
        logging.getLogger("coretexpylib").debug(">> [Coretex] Failed to extract disk read metric")
        return 0

    return float(counters.read_bytes)


def getDiskWrite() -> float:
    """
        Returns
        -------
        float -> total amount of bytes wrote to disk
    """

    counters = psutil.disk_io_counters()

    if counters is None:
        logging.getLogger("coretexpylib").debug(">> [Coretex] Failed to extract disk write metric")
        return 0

    return float(counters.write_bytes)


def getStorageUsage() -> float:
    """
        Returns
        -------
        float -> Used storage space as percentage
    """

    info = shutil.disk_usage("/")
    return info.used / info.total * 100  # 0-1 range -> 0-100 range


def _getNetworkUsage() -> Tuple[float, float]:
    counters = psutil.net_io_counters(pernic = True)

    totalBytesRecv = 0
    totalBytesSent = 0

    for counter in counters.values():
        bytesRecv = counter.bytes_recv
        bytesSent = counter.bytes_sent

        totalBytesRecv += bytesRecv
        totalBytesSent += bytesSent

    return float(totalBytesRecv * 8), float(totalBytesSent * 8)


def getDownloadSpeed() -> float:
    """
        Returns
        -------
        float -> total amount of bytes downloaded
    """

    return _getNetworkUsage()[0]


def getUploadSpeed() -> float:
    """
        Returns
        -------
        float -> total amount of bytes uploaded
    """

    return _getNetworkUsage()[1]



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, List, Dict
from typing_extensions import Self

from .base import BaseObject
from .task import Task
from .project_type import ProjectType
from .project_visibility import ProjectVisibility
from ..entity_visibility_type import EntityVisibilityType
from ...networking import networkManager, NetworkResponse, NetworkRequestError


class Project(BaseObject):

    """
        Represents the project entity from Coretex.ai\n
        Contains properties that describe the project
    """

    tasks: List[Task]
    visibility: ProjectVisibility

    @classmethod
    def createProject(
        cls,
        name: str,
        projectType: ProjectType,
        visiblity: ProjectVisibility = ProjectVisibility.private,
        description: Optional[str] = None
    ) -> Self:
        """
            Creates a new project with the provided name and description

            Parameters
            ----------
            name : str
                project name
            projectType : ProjectType
                type of the created Project
            visibility : ProjectVisibility
                visibility of the created Project
            description : Optional[str]
                project description

            Returns
            -------
            Self -> The created project object

            Raises
            ------
            NetworkRequestError -> If project creation failed

            Example
            -------
            >>> from coretex import Project, ProjectType
            \b
            >>> try:
            >>>     dummyProject = Project.createProject(
                        name = "dummyProject",
                        projectType = ProjectType.other,
                        description = "This is dummy Coretex Project"
                    )
                except:
                    print("Failed to create project.")
        """

        return cls.create(
            name = name,
            project_task = projectType,
            description = description,
            visiblity = visiblity
        )

    @classmethod
    def decode(cls, encodedObject: Dict[str, Any]) -> Self:
        obj = super().decode(encodedObject)
        obj.tasks = Task.fetchAll(queryParameters=[
            f"parentId={obj.id}"
        ])

        return obj

    def addTask(self, name: str, description: Optional[str]) -> bool:
        """
            Adds new task to the project

            Parameters
            ----------
            name : str
                task name
            description : Optional[str]
                task description

            Returns
            -------
            bool -> True if the task was added. False if the task was not added
        """

        task = Task.createTask(name, self.id, description)
        if task is None:
            return False

        self.tasks.append(task)
        return True

    @classmethod
    def fetchByName(cls, name: str) -> Self:
        """
            Fetches Project based on specified name

            Parameters
            ----------
            name : str
                The name of the Project to fetch

            Returns
            -------
            Self -> Fetched Project

            Raises
            ------
            RuntimeError -> If the Project with specified name is not found
        """

        results = cls.fetchAll(name = f"={name}")
        if len(results) == 0:
            raise ValueError(f"Project with name \"{name}\" not found.")

        return results[0]

    def updateVisibility(self, visibility: ProjectVisibility) -> None:
        """
            Updates visibility of the project

            Parameters
            ----------
            visibility : ProjectVisibility
                visibility of the project

            Raises
            ------
            NetworkRequestError -> If request for updating the Project visibility failed
        """

        parameters = {
            "entity_id": self.id,
            "type": EntityVisibilityType.project,
            "visibility": visibility
        }

        response = networkManager.post("entity-visibility", parameters)

        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to update visibility of the Project.")

        self.visibility = visibility



from pathlib import Path

import os

from .node_mode import NodeMode
from ...statistics import getAvailableRam, getAvailableCpuCount


cpuCount = os.cpu_count()

DOCKER_CONTAINER_NAME = "coretex_node"
DOCKER_CONTAINER_NETWORK = "coretex_node"
DEFAULT_STORAGE_PATH = str(Path.home() / ".coretex")
DEFAULT_RAM = getAvailableRam()
MINIMUM_RAM = 6
DEFAULT_SWAP_MEMORY = DEFAULT_RAM * 2
DEFAULT_SHARED_MEMORY = 2
DEFAULT_CPU_COUNT = getAvailableCpuCount()
DEFAULT_NODE_MODE = NodeMode.execution
DEFAULT_ALLOW_DOCKER = False
DEFAULT_NODE_SECRET = ""
DEFAULT_INIT_SCRIPT = ""
DEFAULT_NEAR_WALLET_ID = ""



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional, List, Set, Dict

import os
import json
import glob
import logging

from PIL import Image

from ..base_converter import BaseConverter
from ...annotation import CoretexSegmentationInstance, CoretexImageAnnotation, BBox


class CreateMLConverter(BaseConverter):

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__imagesPath = os.path.join(datasetPath, "images")

        annotations = os.path.join(datasetPath, "annotations")
        self.__fileNames = glob.glob(os.path.join(annotations, "*.json"))

    def _dataSource(self) -> List[str]:
        return self.__fileNames

    def _extractLabels(self) -> Set[str]:
        labels: Set[str] = set()

        for fileName in self.__fileNames:
            with open(fileName) as jsonFile:
                data = json.load(jsonFile)[0]

                for annotation in data["annotations"]:
                    labels.add(annotation["label"])

        return labels

    def __extractBBox(self, bbox: Dict[str, int]) -> BBox:
        return BBox(
            bbox["x"] - bbox["width"] // 2,
            bbox["y"] - bbox["height"] // 2,
            bbox["width"],
            bbox["height"]
        )

    def __extractInstance(self, annotation: Dict[str, Any]) -> Optional[CoretexSegmentationInstance]:
        label = annotation["label"]

        coretexClass = self._dataset.classByName(label)
        if coretexClass is None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] Class: ({label}) is not a part of dataset")
            return None

        bbox = self.__extractBBox(annotation["coordinates"])
        return CoretexSegmentationInstance.create(coretexClass.classIds[0], bbox, [bbox.polygon])

    def __extractImageAnnotation(self, imageAnnotation: Dict[str, Any]) -> None:
        imageName = imageAnnotation["image"]
        image = Image.open(f"{self.__imagesPath}/{imageName}")

        coretexAnnotation = CoretexImageAnnotation.create(imageName, image.width, image.height, [])

        for annotation in imageAnnotation["annotations"]:
            instance = self.__extractInstance(annotation)
            if instance is None:
                continue

            coretexAnnotation.instances.append(instance)

        self._saveImageAnnotationPair(os.path.join(self.__imagesPath, imageName), coretexAnnotation)

    def _extractSingleAnnotation(self, fileName: str) -> None:
        with open(fileName) as jsonFile:
            data = json.load(jsonFile)[0]

            self.__extractImageAnnotation(data)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Dict, Tuple

import xml.etree.ElementTree as ET


def getTag(root: ET.Element, tag: str) -> Optional[str]:
    element = root.find(tag)
    if element is None:
        return None

    return element.text


def toFloat(rootEl: ET.Element, firstEl: str, secondEl: str) -> Tuple[Optional[float], Optional[float]]:
    firstVal = getTag(rootEl, firstEl)
    secondVal = getTag(rootEl, secondEl)

    if firstVal is None or secondVal is None:
        return (None, None)

    return (float(firstVal), float(secondVal))


def toInt(rootEl: ET.Element, firstEl: str, secondEl: str) -> Tuple[Optional[int], Optional[int]]:
    firstVal = getTag(rootEl, firstEl)
    secondVal = getTag(rootEl, secondEl)

    if firstVal is None or secondVal is None:
        return (None, None)

    return (int(firstVal), int(secondVal))


def getBoxes(bndbox: ET.Element) -> Optional[Dict[str, float]]:
    xmin, ymin = toFloat(bndbox, "xmin", "ymin")
    xmax, ymax = toFloat(bndbox, "xmax", "ymax")

    if xmax is None: return None
    if xmin is None: return None
    if ymax is None: return None
    if ymin is None: return None

    return {
        "top_left_x": xmin,
        "top_left_y": ymin,
        "width": xmax - xmin,
        "height": ymax - ymin,
    }


#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any, Optional
from pathlib import Path

import os
import json
import sys


def isCliRuntime() -> bool:
    executablePath = sys.argv[0]
    return (
        executablePath.endswith("/bin/coretex") and
        os.access(executablePath, os.X_OK)
    )


def getEnvVar(key: str, default: str) -> str:
    if os.environ.get(key) is None:
        return default

    return os.environ[key]


CONFIG_DIR = Path.home().joinpath(".config", "coretex")
DEFAULT_CONFIG_PATH = CONFIG_DIR / "config.json"


DEFAULT_CONFIG: Dict[str, Any] = {
    # os.environ used directly here since we don't wanna
    # set those variables to any value if they don't exist
    # in the os.environ the same way we do for properties which
    # call genEnvVar
    "username": os.environ.get("CTX_USERNAME"),
    "password": os.environ.get("CTX_PASSWORD"),
    "token": None,
    "refreshToken": None,
    "serverUrl": getEnvVar("CTX_API_URL", "https://api.coretex.ai/"),
    "storagePath": getEnvVar("CTX_STORAGE_PATH", "~/.coretex")
}


def loadConfig() -> Dict[str, Any]:
    with DEFAULT_CONFIG_PATH.open("r") as configFile:
        try:
            config: Dict[str, Any] = json.load(configFile)
        except json.JSONDecodeError:
            config = {}

    for key, value in DEFAULT_CONFIG.items():
        if not key in config:
            config[key] = value

    return config


def _syncConfigWithEnv() -> None:
    # If configuration does not exist create default one
    if not DEFAULT_CONFIG_PATH.exists():
        DEFAULT_CONFIG_PATH.parent.mkdir(parents = True, exist_ok = True)
        config = DEFAULT_CONFIG.copy()
    else:
        config = loadConfig()

    saveConfig(config)

    if not "CTX_API_URL" in os.environ:
        os.environ["CTX_API_URL"] = config["serverUrl"]

    secretsKey = config.get("secretsKey")
    if isinstance(secretsKey, str) and secretsKey != "":
        os.environ["CTX_SECRETS_KEY"] = secretsKey

    if not isCliRuntime():
        os.environ["CTX_STORAGE_PATH"] = config["storagePath"]
    else:
        os.environ["CTX_STORAGE_PATH"] = f"{CONFIG_DIR}/data"


def saveConfig(config: Dict[str, Any]) -> None:
    configPath = DEFAULT_CONFIG_PATH.expanduser()
    with configPath.open("w+") as configFile:
        json.dump(config, configFile, indent = 4)


def isUserConfigured(config: Dict[str, Any]) -> bool:
    return (
        config.get("username") is not None and
        config.get("password") is not None and
        config.get("storagePath") is not None
    )


def isNodeConfigured(config: Dict[str, Any]) -> bool:
    return (
        config.get("nodeName") is not None and
        config.get("storagePath") is not None and
        config.get("image") is not None and
        config.get("serverUrl") is not None and
        config.get("nodeAccessToken") is not None and
        config.get("nodeRam") is not None and
        config.get("nodeSwap") is not None and
        config.get("nodeSharedMemory") is not None and
        config.get("nodeMode") is not None
    )


def getInitScript(config: Dict[str, Any]) -> Optional[Path]:
    value = config.get("initScript")

    if not isinstance(value, str):
        return None

    if value == "":
        return None

    path = Path(value).expanduser().absolute()
    if not path.exists():
        return None

    return path



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .int_parameter import IntParameter
from .float_parameter import FloatParameter
from .str_parameter import StrParameter
from .bool_parameter import BoolParameter
from .dataset_parameter import DatasetParameter
from .model_parameter import ModelParameter
from .imu_vectors_parameter import IMUVectorsParameter
from .enum_parameter import EnumParameter
from .range_parameter import RangeParameter
from .secret_parameter import SecretParameter
from .list_int_parameter import ListIntParameter
from .list_float_parameter import ListFloatParameter
from .list_str_parameter import ListStrParameter
from .list_dataset_parameter import ListDatasetParameter
from .list_model_parameter import ListModelParameter
from .list_enum_parameter import ListEnumParameter
from .list_secret_parameter import ListSecretParameter



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypeVar, Generic, Union
from abc import ABC, abstractmethod
from zipfile import BadZipFile, ZipFile
from pathlib import Path

import shutil


SampleDataType = TypeVar("SampleDataType")


class Sample(ABC, Generic[SampleDataType]):

    """
        Represents the generic class Sample
        Includes methods that can be used by any instance of Sample
        and abstract methods that must be implemented by any subclass
    """

    name: str

    @property
    @abstractmethod
    def path(self) -> Path:
        pass

    @property
    @abstractmethod
    def zipPath(self) -> Path:
        pass

    @abstractmethod
    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        """
            Downloads the Sample if it is an instance or a subclass of NetworkSample
            Ignored for instances and subclasses of LocalSample
        """

        pass

    def __unzipSample(self) -> None:
        if self.path.exists():
            shutil.rmtree(self.path)

        with ZipFile(self.zipPath) as zipFile:
            zipFile.extractall(self.path)

    def unzip(self, ignoreCache: bool = False) -> None:
        """
            Unzip sample zip file

            Parameters
            ----------
            ignoreCache : bool
                if set to false performs unzip action even if
                sample is previously unzipped
        """

        if not self.zipPath.exists():
            raise FileNotFoundError(f"Sample \"{self.name}\" archive does not exist at path \"{self.zipPath}\"")

        if self.path.exists() and not ignoreCache:
            return

        try:
            self.__unzipSample()
        except BadZipFile:
            # Delete invalid zip file
            self.zipPath.unlink()

            # Re-download
            self.download()

            # Try to unzip - if it fails again it should crash
            self.__unzipSample()

    @abstractmethod
    def load(self) -> SampleDataType:
        pass

    def joinPath(self, other: Union[Path, str]) -> Path:
        """
            Joins sample path and provided path

            Parameters
            ----------
            other : Union[Path, str]
                path for join

            Returns
            -------
            Path -> path created from sample path and provided path

            Example
            -------
            >>> print(sampleObj.joinPath("dummy.zip"))
            Path("path/to/sample/dummy.zip")
        """

        if isinstance(other, str):
            other = Path(other)

        return self.path / other

    def _updateArchive(self) -> None:
        oldZipPath = self.zipPath.parent / f"{self.zipPath.stem}-old.zip"
        self.zipPath.rename(oldZipPath)

        with ZipFile(self.zipPath, "w") as zipFile:
            for value in self.path.rglob("*"):
                if not value.is_file():
                    continue

                zipFile.write(value, value.relative_to(self.path))

        oldZipPath.unlink()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, List, Dict, Union, Tuple, TypeVar, Generic, Type
from typing_extensions import Self, override
from zipfile import ZipFile, ZIP_DEFLATED
from pathlib import Path

import os
import time
import logging
import zipfile
import json

from .utils import createSnapshot
from .artifact import Artifact
from .status import TaskRunStatus
from .metrics import Metric, MetricType
from .parameter import validateParameters, parameter_factory
from .execution_type import ExecutionType
from ..dataset import Dataset, LocalDataset, NetworkDataset
from ..project import ProjectType
from ..model import Model
from ... import folder_manager
from ...codable import KeyDescriptor
from ...networking import networkManager, NetworkObject, NetworkRequestError, FileData


DatasetType = TypeVar("DatasetType", bound = Dataset)

class TaskRun(NetworkObject, Generic[DatasetType]):

    """
        Represents TaskRun entity from Coretex.ai

        Properties
        ----------
        datasetId : int
            id of dataset
        name : str
            name of TaskRun
        description : str
            description of TaskRun
        meta : Dict[str, Any]
            meta data of TaskRun
        status : TaskRunStatus
            status of TaskRun
        projectId : int
            id of Coretex Project
        projectName : str
            name of Coretex Project
        projectType : ProjectType
            appropriate project type
        taskId : int
            id of task
        taskName : str
            name of task
        createdById : str
            id of created TaskRun
        useCachedEnv : bool
            if True chached env will be used, otherwise new environment will be created
    """

    name: str
    description: str
    meta: Dict[str, Any]
    status: TaskRunStatus
    projectId: int
    projectName: str
    projectType: ProjectType
    taskId: int
    taskName: str
    entryPoint: str
    createdById: str
    useCachedEnv: bool
    executionType: ExecutionType
    metrics: List[Metric]

    def __init__(self) -> None:
        super(TaskRun, self).__init__()

        self.metrics = []
        self.__parameters: Dict[str, Any] = {}

    @property
    def parameters(self) -> Dict[str, Any]:
        """
            Returns
            -------
            Dict[str, Any] -> Parameters for TaskRun
        """

        return self.__parameters

    @property
    def taskPath(self) -> Path:
        """
            Returns
            -------
            Path -> Path for TaskRun
        """

        return folder_manager.temp / str(self.id)

    @property
    def dataset(self) -> DatasetType:
        """
            Value of the parameter with name "dataset" assigned to this TaskRun

            Returns
            -------
            Dataset object if there was a parameter with name "dataset" entered when the TaskRun was started

            Raises
            ------
            ValueError -> if there is not parameter with name "dataset"
        """

        dataset = self.parameters.get("dataset")
        if dataset is None:
            raise ValueError(f">> [Coretex] TaskRun \"{self.id}\" does not have a parameter named \"dataset\"")

        return dataset  # type: ignore

    @property
    def isLocal(self) -> bool:
        return self.executionType == ExecutionType.local

    def setDatasetType(self, datasetType: Type[DatasetType]) -> None:
        for key, value in self.__parameters.items():
            if isinstance(value, LocalDataset) and issubclass(datasetType, LocalDataset):
                self.__parameters[key] = datasetType(value.path)  # type: ignore

            if isinstance(value, NetworkDataset) and issubclass(datasetType, NetworkDataset):
                self.__parameters[key] = datasetType.fetchById(value.id)

    def setModelType(self, modelType: Type[Model]) -> None:
        for key, value in self.__parameters.items():
            if isinstance(value, Model):
                self.__parameters[key] = modelType.fetchById(value.id)

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["status"] = KeyDescriptor("status", TaskRunStatus)
        descriptors["projectId"] = KeyDescriptor("project_id")
        descriptors["projectName"] = KeyDescriptor("project_name")
        descriptors["projectType"] = KeyDescriptor("project_task", ProjectType)
        descriptors["taskId"] = KeyDescriptor("sub_project_id")
        descriptors["taskName"] = KeyDescriptor("sub_project_name")
        descriptors["executionType"] = KeyDescriptor("execution_type", ExecutionType)

        # private properties of the object should not be encoded
        descriptors["__parameters"] = KeyDescriptor(isEncodable = False)

        return descriptors

    @classmethod
    @override
    def _endpoint(cls) -> str:
        return "model-queue"

    @override
    def entityUrl(self) -> str:
        return f"run?id={self.id}"

    def onDecode(self) -> None:
        super().onDecode()

        if self.meta["parameters"] is None:
            self.meta["parameters"] = []

        if not isinstance(self.meta["parameters"], list):
            raise ValueError(">> [Coretex] Invalid parameters")

        parameters = [parameter_factory.create(value) for value in self.meta["parameters"]]

        parameterValidationResults = validateParameters(parameters, verbose = True)
        if not all(parameterValidationResults.values()):
            raise ValueError("Invalid parameters found")

        self.__parameters = {parameter.name: parameter.parseValue(self.projectType) for parameter in parameters}

    def _isInterval(self, metricName: str) -> bool:
        for metric in self.metrics:
            if metric.name == metricName and metric.xType == MetricType.interval.name:
                return True

        return False

    def updateStatus(
        self,
        status: Optional[TaskRunStatus] = None,
        message: Optional[str] = None,
        notifyServer: bool = True
    ) -> bool:

        """
            Updates TaskRun status, if message parameter is None
            default message value will be used\n
            Some TaskRun statuses do not have default message

            Parameters
            ----------
            status : Optional[TaskRunStatus]
                Status to which the TaskRun will be updated to
            message : Optional[str]
                Descriptive message for TaskRun status, it is diplayed
                when the status is hovered on the Coretex Web App
            notifyServer : bool
                if True update request will be sent to Coretex.ai

            Example
            -------
            >>> from coretex import ExecutingTaskRun, TaskRunStatus
            \b
            >>> ExecutingTaskRun.current().updateStatus(
                    TaskRunStatus.completedWithSuccess
                )
            True
        """

        if status is not None:
            self.status = status

        if notifyServer:
            if status is not None and message is None:
                message = status.defaultMessage

            parameters: Dict[str, Any] = {
                "id": self.id
            }

            if status is not None:
                parameters["status"] = status

            if message is not None:
                parameters["message"] = message

            # TODO: Should API rename this too?
            endpoint = f"{self._endpoint()}/job-status-update"
            response = networkManager.post(endpoint, parameters)

            if response.hasFailed():
                logging.getLogger("coretexpylib").error(">> [Coretex] Error while updating TaskRun status")
            elif status is not None:
                logging.getLogger("coretexpylib").info(f">> [Coretex] Updated Task Run status to \"{status.name}\"")

            return not response.hasFailed()

        return True

    def createMetrics(self, metrics: List[Metric]) -> None:
        """
            Creates specified metrics for the TaskRun

            Parameters
            ----------
            values : List[Metric]]
                List of Metric meta objects in this format
                Metric("name", "x_label", "x_type", "y_label", "y_type", "x_range", "y_range")

            Returns
            -------
            List[Metric] -> List of Metric objects

            Raises
            ------
            NetworkRequestError -> if the request failed

            Example
            -------
            >>> from coretex import ExecutingTaskRun, MetricType
            \b
            >>> metrics = ExecutingTaskRun.current().createMetrics([
                    Metric.create("loss", "epoch", MetricType.int, "value", MetricType.float, None, [0, 100]),
                    Metric.create("accuracy", "epoch", MetricType.int, "value", MetricType.float, None, [0, 100])
                ])
            >>> if len(metrics) == 0:
                    print("Failed to create metrics")
        """

        parameters: Dict[str, Any] = {
            "experiment_id": self.id,
            "metrics": [metric.encode() for metric in metrics]
        }

        response = networkManager.post(f"{self._endpoint()}/metrics-meta", parameters)
        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to create metrics")

        self.metrics.extend(metrics)

    def submitMetrics(self, metricValues: Dict[str, Tuple[float, float]]) -> bool:
        """
            Appends metric values for the provided metrics

            Parameters
            ----------
            metricValues : Dict[str, Tuple[float, float]]
                Values of metrics in this format {"name": x, y}

            Example
            -------
            >>> from coretex import ExecutingTaskRun
            \b
            >>> result = ExecutingTaskRun.current().submitMetrics({
                    "loss": (epoch, logs["loss"]),
                    "accuracy": (epoch, logs["accuracy"]),
                })
            >>> print(result)
            True
        """

        metrics = [{
            "timestamp": value[0] if self._isInterval(key) else time.time(),
            "metric": key,
            "x": value[0],
            "y": value[1]
        } for key, value in metricValues.items()]

        parameters: Dict[str, Any] = {
            "experiment_id": self.id,
            "metrics": metrics
        }

        response = networkManager.post(f"{self._endpoint()}/metrics", parameters)
        return not response.hasFailed()

    def submitOutput(self, parameterName: str, value: Any) -> None:
        """
            Submit an output of this task to act as a parameter in tasks
            downstream in the Workflow

            Parameters
            ----------
            parameterName : str
                name of the parameter
            value : Any
                value that will be sent (id of the object will be sent
                in case Coretex objects like CustomDataset, Model etc. is submited)

            Raises
            ------
            NetworkRequestError -> if the request failed

            Example
            -------
            >>> from coretex import TaskRun
            \b
            >>> ExecutingTaskRun.current().submitOutputParameter(
                    parameterName = "outputDataset"
                    value = outputDataset
                )
        """

        self.submitOutputs({parameterName: value})

    def submitOutputs(self, outputs: Dict[str, Any]) -> None:
        """
            Submit multiple outputs of this task to act as parameters in tasks
            downstream in the Workflow

            Parameters
            ----------
            outputs : Dict[str, Any]
                dictionary with outputs, with key being the name of the parameter
                (id of the object will be sent in case Coretex objects like
                CustomDataset, Model etc. is submited)

            Raises
            ------
            NetworkRequestError -> if the request failed

            Example
            -------
            >>> from coretex import TaskRun
            \b
            >>> result = ExecutingTaskRun.current().submitOutputParameters({
                    "outputDataset": outputDataset,
                    "numbers": 123
                })
        """

        outputParameters: List[Dict[str, Any]] = []

        for key, value in outputs.items():
            if isinstance(value, NetworkObject):
                value = value.id

            outputParameters.append({key: value})

        parameters: Dict[str, Any] = {
            "id": self.id,
            "parameters": outputParameters
        }

        response = networkManager.post(f"{self._endpoint()}/output-parameter", parameters)
        if response.hasFailed():
            raise NetworkRequestError(response, ">> [Coretex] Failed to submit outputs")

    def downloadTask(self) -> bool:
        """
            Downloads task snapshot linked to the TaskRun

            Returns
            -------
            bool -> True if task downloaded successfully, False if task download has failed
        """

        params = {
            "model_queue_id": self.id
        }

        zipFilePath = f"{self.taskPath}.zip"
        response = networkManager.download(f"workspace/download", zipFilePath, params)

        if response.hasFailed():
            logging.getLogger("coretexpylib").info(">> [Coretex] Task download has failed")
            return False

        with ZipFile(zipFilePath) as zipFile:
            zipFile.extractall(self.taskPath)

        # remove zip file after extract
        os.unlink(zipFilePath)

        return not response.hasFailed()

    def createArtifact(
        self,
        localFilePath: Union[Path, str],
        remoteFilePath: str,
        mimeType: Optional[str] = None
    ) -> Optional[Artifact]:

        """
            Creates Artifact for the current TaskRun on Coretex.ai

            Parameters
            ----------
            localFilePath : Union[Path, str]
                local path of Artifact file
            remoteFilePath : str
                path of Artifact file on Coretex
            mimeType : Optional[str]
                mimeType (not required) if not passed guesMimeType() function is used

            Returns
            -------
            Optional[Artifact] -> if response is True returns Artifact object, None otherwise
        """

        return Artifact.create(self.id, localFilePath, remoteFilePath, mimeType)

    def createQiimeArtifact(self, rootArtifactFolderName: str, qiimeArtifactPath: Path) -> None:
        if not zipfile.is_zipfile(qiimeArtifactPath):
            raise ValueError(">> [Coretex] Not an archive")

        localFilePath = str(qiimeArtifactPath)
        remoteFilePath = f"{rootArtifactFolderName}/{qiimeArtifactPath.name}"

        mimeType: Optional[str] = None
        if qiimeArtifactPath.suffix in [".qza", ".qzv"]:
            mimeType = "application/zip"

        artifact = self.createArtifact(localFilePath, remoteFilePath, mimeType)
        if artifact is None:
            logging.getLogger("coretexpylib").warning(f">> [Coretex] Failed to upload {localFilePath} to {remoteFilePath}")

        # TODO: Enable when uploading file by file is not slow anymore
        # tempDir = Path(FolderManager.instance().createTempFolder(rootArtifactFolderName))
        # fileUtils.recursiveUnzip(qiimeArtifactPath, tempDir, remove = False)

        # for path in fileUtils.walk(tempDir):
        #     relative = path.relative_to(tempDir)

        #     localFilePath = str(path)
        #     remoteFilePath = f"{rootArtifactFolderName}/{str(relative)}"

        #     logging.getLogger("coretexpylib").debug(f">> [Coretex] Uploading {localFilePath} to {remoteFilePath}")

        #     artifact = self.createArtifact(localFilePath, remoteFilePath)
        #     if artifact is None:
        #         logging.getLogger("coretexpylib").warning(f">> [Coretex] Failed to upload {localFilePath} to {remoteFilePath}")

    @classmethod
    def run(
        cls,
        taskId: int,
        nodeId: Union[int, str],
        name: Optional[str],
        description: Optional[str] = None,
        parameters: Optional[List[Dict[str, Any]]] = None
    ) -> Self:

        """
            Schedules an TaskRun for execution on the specified
            Node on Coretex.ai

            Parameters
            ----------
            taskId : int
                id of task that is being used for starting TaskRun
            nodeId : Union[int, str]
                id of node that is being used for starting TaskRun
            name : Optional[str]
                name of TaskRun (not required)
            description : Optional[str]
                TaskRun description (not required)
            parameters : Optional[List[Dict[str, Any]]]
                list of parameters (not required)

            Returns
            -------
            Self -> TaskRun object

            Raises
            ------
            NetworkRequestError -> if the request failed

            Example
            -------
            >>> from coretex import TaskRun
            >>> from coretex.networking import NetworkRequestError
            \b
            >>> parameters = [
                    {
                        "name": "dataset",
                        "description": "Dataset id that is used for fetching dataset from coretex.",
                        "value": null,
                        "data_type": "dataset",
                        "required": true
                    }
                ]
            \b
            >>> try:
                    taskRun = TaskRun.run(
                        taskId = 1023,
                        nodeId = 23,
                        name = "Dummy Custom TaskRun
                        description = "Dummy description",
                        parameters = parameters
                    )

                    print(f"Created TaskRun with name: {taskRun.name}")
            >>> except NetworkRequestError:
                    print("Failed to create TaskRun")
        """

        if isinstance(nodeId, int):
            nodeId = str(nodeId)

        if parameters is None:
            parameters = []

        response = networkManager.post("run", {
            "sub_project_id": taskId,
            "service_id": nodeId,
            "name": name,
            "description": description,
            "execution_type": ExecutionType.remote.value,
            "parameters": parameters
        })

        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to create TaskRun")

        responseJson = response.getJson(dict)
        return cls.fetchById(responseJson["experiment_ids"][0])

    @classmethod
    def runLocal(
        cls,
        projectId: int,
        saveSnapshot: bool,
        name: Optional[str],
        description: Optional[str] = None,
        parameters: Optional[List[Dict[str, Any]]] = None,
        entryPoint: Optional[str] = None
    ) -> Self:

        """
            Creates TaskRun on Coretex.ai with the provided parameters,
            which will be run on the same machine which created it immidiately
            after running the entry point file of the task

            Parameters
            ----------
            projectId : int
                id of project that is being used for starting TaskRun
            saveSnapshot : bool
                true if snapshot of local files will be saved to Coretex
            name : Optional[str]
                name of TaskRun (not required)
            description : Optional[str]
                TaskRun description (not required)
            parameters : Optional[List[Dict[str, Any]]]
                list of parameters (not required)
            entryPoint : Optional[str]
                relative path to the script inside of the project

            Returns
            -------
            Self -> TaskRun object

            Raises
            ------
            NetworkRequestError -> if the request failed
        """

        if parameters is None:
            parameters = []

        params = {
            "project_id": projectId,
            "name": name,
            "description": description,
            "execution_type": ExecutionType.local.value,
            "parameters": json.dumps(parameters)
        }

        if entryPoint is not None:
            params["entry_point"] = entryPoint

        # Create snapshot
        if saveSnapshot:
            files = [FileData.createFromPath("file", createSnapshot())]
        else:
            files = None

        response = networkManager.formData("run", params, files)
        if response.hasFailed():
            raise NetworkRequestError(response, "Failed to create TaskRun")

        responseJson = response.getJson(dict)
        return cls.fetchById(responseJson["experiment_ids"][0])



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

import logging

import click

from .ui import errorEcho


class ClickExceptionInterceptor(click.Group):

    def invoke(self, ctx: click.Context) -> Any:
        try:
            return super().invoke(ctx)
        except BaseException as exc:
            if isinstance(exc, KeyboardInterrupt):
                raise

            if isinstance(exc, click.exceptions.Exit) and exc.exit_code == 0:
                raise

            self.handleException(ctx, exc)

    def handleException(self, ctx: click.Context, exc: BaseException) -> None:
        click.echo(click.style(str(exc), fg = "red"))
        logging.getLogger("cli").debug(exc, exc_info = exc)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, TypeVar, Generic, List, Callable
from abc import ABC, abstractmethod
from pathlib import Path

from ..sample import Sample


SampleType = TypeVar("SampleType", bound = "Sample")


class Dataset(ABC, Generic[SampleType]):

    """
        Represents the generic class Dataset
        Includes methods that can be used by any instance of Dataset
        and abstract methods that must be implemented by any subclass

        Properties
        ----------
        name : str
            name of dataset
        samples : List[SampleType]
            list of samples
    """

    name: str
    samples: List[SampleType]

    @property
    def count(self) -> int:
        """
            Returns
            -------
            int -> number of samples in this dataset
        """

        return len(self.samples)

    @property
    @abstractmethod
    def path(self) -> Path:
        pass

    @abstractmethod
    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        pass

    def rename(self, name: str) -> bool:
        """
            Renames the dataset, if the provided name is
            different from the current name

            Parameters
            ----------
            name : str
                new dataset name

            Returns
            -------
            bool -> True if dataset was renamed, False if dataset was not renamed
        """

        if self.name == name:
            return False

        self.name = name
        return True

    def getSample(self, name: str) -> Optional[SampleType]:
        """
            Retrieves sample which matches the provided name

            Parameters
            ----------
            name : str
                name of sample

            Returns
            -------
            Optional[SampleType] -> sample object
        """

        for sample in self.samples:
            # startswith must be used since if we import sample
            # with the same name twice, the second one will have
            # suffix with it's serial number
            if sample.name.startswith(name):
                return sample

        return None

    def getSamples(self, filterFunc: Callable[[SampleType], bool]) -> List[SampleType]:
        filteredSamples: List[SampleType] = []

        for sample in self.samples:
            if filterFunc(sample):
                filteredSamples.append(sample)

        return filteredSamples



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import List, Iterator
from pathlib import Path
from contextlib import contextmanager

import logging

from watchdog.events import FileSystemEventHandler, FileSystemEvent
from watchdog.observers import Observer

from ...entities import TaskRun


IGNORED_FILES = ["_coretex.py"]


class FileEventHandler(FileSystemEventHandler):

    def __init__(self) -> None:
        super().__init__()

        self.artifactPaths: List[Path] = []

    def on_created(self, event: FileSystemEvent) -> None:
        if event.is_directory:
            return

        filePath = Path(event.src_path)

        if filePath.parent.joinpath(".coretexignore").exists():
            return

        if filePath.name in IGNORED_FILES:
            return

        logging.getLogger("coretex").debug(f">> [Coretex] File created at path \"{filePath}\", adding to artifacts list")
        self.artifactPaths.append(filePath)


@contextmanager
def track(taskRun: TaskRun) -> Iterator[FileEventHandler]:
    # If local use current working dir, else use task path
    root = Path.cwd() if taskRun.isLocal else taskRun.taskPath

    try:
        observer = Observer()
        observer.setName("ArtifactTracker")

        logging.getLogger("coretexpylib").debug(f">> [Coretex] Tracking files created inside \"{root}\"")

        eventHandler = FileEventHandler()
        observer.schedule(eventHandler, root, recursive = True)  # type: ignore[no-untyped-call]
        try:
            observer.start()  # type: ignore[no-untyped-call]
        except OSError as e:
            logging.getLogger("coretexpylib").error(f"Failed to start watchdog observer with error: {e}")

        yield eventHandler
    finally:
        observer.stop()  # type: ignore[no-untyped-call]
        observer.join()

        for index, artifactPath in enumerate(eventHandler.artifactPaths):
            logging.getLogger("coretexpylib").debug(f">> [Coretex] Uploading {index + 1}/{len(eventHandler.artifactPaths)} - \"{artifactPath}\"")

            try:
                artifact = taskRun.createArtifact(artifactPath, str(artifactPath.relative_to(root)))
                if artifact is not None:
                    logging.getLogger("coretexpylib").debug(f"\tSuccessfully uploaded artifact")
                else:
                    logging.getLogger("coretexpylib").debug(f"\tFailed to upload artifact")
            except Exception as e:
                logging.getLogger("coretexpylib").error(f"\tError while creating artifact: {e}")
                logging.getLogger("coretexpylib").debug(f"\tError while creating artifact: {e}", exc_info = e)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple
from pathlib import Path

import subprocess
import os

from ...utils import command, logProcessOutput, CommandException
from ...entities import CustomDataset
from ...logging import LogSeverity


def indexCommand(bwaPath: Path, sequencePath: Path, prefix: Path) -> None:
    """
        This function acts as a wrapper for the index command of BWA
        (Burrows-Wheeler Aligner). It is necessary to perform indexing
        on the reference sequence before alignment.

        Parameters
        ----------
        bwaPath : Path
            Path pointing to the bwa binary
        sequencePath : Path
            Path pointing to the reference sequence
        prefix : Path
            Path serving as the prefix for indexing. This command will
            generate 5 new files that are needed for alignment. These files will
            have the path "{prefix}.{suffix}", each with a different suffix / extension

        Example
        -------
        >>> from pathlib import Path
        >>> bwaPath = Path("tools/bwa")
        >>> sequencePath = Path("sequences/hg18.fa")
        >>> prefix = Path("output/hg18")
        >>> indexCommand(bwaPath, sequencePath, prefix)

        Link to BWA: https://bio-bwa.sourceforge.net
    """

    command([
        str(bwaPath.absolute()), "index",
        "-p", str(prefix.absolute()),
        str(sequencePath.absolute())
    ])


def alignCommand(
    bwaPath: Path,
    prefix: Path,
    sequencePath: Path,
    outputPath: Path,
    multithreading: bool = True
) -> None:

    """
        This function acts as a wrapper for the mem command of BWA
        (Burrows-Wheeler Aligner). It perfoms alignment of a given sequence read
        to the reference sequence and outputs a SAM file.

        Parameters
        ----------
        bwaPath : Path
            Path pointing to the bwa binary
        prefix : Path
            Path that serves as the prefix for the 5 files generated during indexing
        sequencePath : Path
            Path pointing to the sequence read file on which alignment should be performed
        outputPath : Path
            Path of the output file

        Example
        -------
        >>> from pathlib import Path
        >>> bwaPath = Path("tools/bwa")
        >>> prefix = Path("reference/hg18")
        >>> sequencePath = Path("input/R34D.fastq")
        >>> outputPath = Path("output/R34D.sam")
        >>> alignCommand(bwaPath, prefix, sequencePath, outputPath)

        Link to BWA: https://bio-bwa.sourceforge.net
    """

    args = [
        str(bwaPath.absolute()), "mem",
        "-o", str(outputPath.absolute())
    ]

    if multithreading:
        threads = os.cpu_count()
        if threads is not None:
            args.extend(["-t", str(threads)])

    args.extend([
        str(prefix.absolute()),
        str(sequencePath.absolute())
    ])

    command(args, True)


def sam2bamCommand(
    samtoolsPath: Path,
    samPath: Path,
    outputPath: Path,
    multithreading: bool = True
) -> None:

    """
        This function uses the CLI tool "samtools" to convert SAM files into their binary
        version, BAM.

        Parameters
        ----------
        samtoolsPath : Path
            Path pointing to the samtools binary
        samPath : Path
            Path pointing to the input .sam file
        outputPath : Path
            Path pointing to the output, .bam, file

        Example
        -------
        >>> from pathlib import Path
        >>> samtoolsPath = Path("tools/samtools")
        >>> samPath = Path("input/R34D.sam")
        >>> outputPath = Path("output/R34D.bam")
        >>> sam2bamCommand(samtoolsPath, samPath, outputPath)

        Link to samtools: http://htslib.org/
    """

    args = [
        str(samtoolsPath.absolute()), "view",
        "-b", "-S"
    ]

    if multithreading:
        threads = os.cpu_count()
        if threads is not None:
            args.extend(["--threads", str(threads - 1)])

    args.extend([
        "-o", str(outputPath.absolute()),
        str(samPath.absolute())
    ])

    command(args)


def extractData(samtoolsPath: Path, file: Path) -> Tuple[List[int], List[int], List[int]]:
    """
        Takes an aligned sequence file (SAM/BAM) and returns three lists holding
        MAPQ scores, leftmost position and sequence length for each read in the
        file. This is done using the samtools view command.

        Parameters
        ----------
        samtoolsPath : Path
            Path pointing to the samtools binary
        samPath : Path
            Path pointing to the input .sam or .bam file

        Example
        -------
        >>> from pathlib import Path
        >>> samtoolsPath = Path("tools/samtools")
        >>> file = Path("R34D.bam")
        >>> scores, positions, lengths = extractData(samtoolsPath, file)

        Link to samtools: http://htslib.org/
    """

    scores: List[int] = []
    positions: List[int] = []
    sequenceLengths: List[int] = []

    args = [
        str(samtoolsPath.absolute()),
        "view",
        "-F", "256", "-F", "2048", "-F", "512",  # Will ignore all duplicate reads
        str(file.absolute())
    ]

    process = subprocess.Popen(
        args,
        shell = False,
        cwd = Path(__file__).parent,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    while process.poll() is None:
        if process.stdout is None:
            continue

        stdout = process.stdout.readlines()

        for line in stdout:
            if len(line) > 0:
                fields = line.decode("UTF-8").strip().split("\t")
                scores.append(int(fields[4]))
                positions.append(int(fields[3]))
                sequenceLengths.append(len(fields[9]))

    if process.stderr is not None:
        for stderr in process.stderr:
            if len(stderr) > 0:
                if process.returncode == 0:
                    logProcessOutput(stderr, LogSeverity.warning)
                else:
                    logProcessOutput(stderr, LogSeverity.fatal)

    if process.returncode != 0:
        raise CommandException(f">> [Coretex] Falied to execute command. Returncode: {process.returncode}")

    return scores, positions, sequenceLengths


def chmodX(file: Path) -> None:
    command(["chmod", "+x", str(file.absolute())])


def loadFa(dataset: CustomDataset) -> List[Path]:
    inputFiles: List[Path] = []

    for sample in dataset.samples:
        sample.unzip()

        for file in Path(sample.path).iterdir():
            if file.suffix.startswith(".fa"):
                inputFiles.append(file)

    if len(inputFiles) == 0:
        raise ValueError(">> [Coretex] No sequence reads found")

    return inputFiles
